"""Schema utils for interfacing with the duckdb database."""

from typing import Optional, cast

from .schema import (
  BINARY,
  BOOLEAN,
  DATE,
  EMBEDDING,
  FLOAT16,
  FLOAT32,
  FLOAT64,
  INT8,
  INT16,
  INT32,
  INT64,
  INTERVAL,
  NULL,
  ROWID,
  SPAN_KEY,
  STRING,
  STRING_SPAN,
  TEXT_SPAN_END_FEATURE,
  TEXT_SPAN_START_FEATURE,
  TIME,
  TIMESTAMP,
  UINT8,
  UINT16,
  UINT32,
  UINT64,
  VALUE_KEY,
  DataType,
  Field,
  MapType,
  Schema,
)


def escape_string_literal(string: str) -> str:
  """Escape a string literal for duckdb."""
  string = string.replace("'", "''")
  return f"'{string}'"


def escape_col_name(col_name: str) -> str:
  """Escape a column name for duckdb."""
  col_name = col_name.replace('"', '""')
  return f'"{col_name}"'


def _duckdb_type(dtype: Optional[DataType]) -> str:
  """Convert the dtype to a duckdb type."""
  if dtype == STRING:
    return 'VARCHAR'
  elif dtype == BOOLEAN:
    return 'BOOLEAN'
  elif dtype == FLOAT16:
    return 'FLOAT'
  elif dtype == FLOAT32:
    return 'FLOAT'
  elif dtype == FLOAT64:
    return 'DOUBLE'
  elif dtype == INT8:
    return 'TINYINT'
  elif dtype == INT16:
    return 'SMALLINT'
  elif dtype == INT32:
    return 'INTEGER'
  elif dtype == INT64:
    return 'BIGINT'
  elif dtype == UINT8:
    return 'UTINYINT'
  elif dtype == UINT16:
    return 'USMALLINT'
  elif dtype == UINT32:
    return 'UINTEGER'
  elif dtype == UINT64:
    return 'UBIGINT'
  elif dtype == BINARY:
    return 'BLOB'
  elif dtype == TIME:
    return 'TIME'
  elif dtype == DATE:
    return 'DATE'
  elif dtype == TIMESTAMP:
    return 'TIMESTAMP'
  elif dtype == INTERVAL:
    return 'INTERVAL'
  elif dtype == EMBEDDING:
    # We reserve an empty column for embeddings in parquet files so they can be queried.
    # The values are *not* filled out. If parquet and duckdb support embeddings in the future, we
    # can set this dtype to the relevant type.
    return 'NULL'
  elif dtype == STRING_SPAN:
    return (
      f'STRUCT("{SPAN_KEY}" STRUCT("{TEXT_SPAN_START_FEATURE}" INT, "{TEXT_SPAN_END_FEATURE}" INT))'
    )
  elif dtype == NULL:
    return 'NULL'
  elif dtype is None:
    return 'NULL'
  elif dtype.type == 'map':
    map_dtype = cast(MapType, dtype)
    return f'MAP({_duckdb_type(map_dtype.key_type)}, {_duckdb_struct(map_dtype.value_field)})'
  else:
    raise ValueError(f'Can not convert dtype "{dtype}" to duckdb type')


def _duckdb_struct(field: Field) -> str:
  """Convert a field to a duckdb struct."""
  if field.fields:
    fields: dict[str, str] = {}
    for name, subfield in field.fields.items():
      fields[name] = _duckdb_type(subfield.dtype) if name == ROWID else _duckdb_struct(subfield)

    # When nodes have both dtype and children, we add __value__ alongside the fields.
    if field.dtype:
      if field.dtype == STRING_SPAN:
        fields[SPAN_KEY] = f'STRUCT("{TEXT_SPAN_START_FEATURE}" INT, "{TEXT_SPAN_END_FEATURE}" INT)'
      else:
        fields[VALUE_KEY] = _duckdb_type(field.dtype)

    members = [f'{escape_col_name(name)} {dtype}' for name, dtype in fields.items()]
    return f'STRUCT({", ".join(members)})'

  if field.repeated_field:
    return f'{_duckdb_struct(field.repeated_field)}[]'

  return _duckdb_type(field.dtype)


def duckdb_schema(schema: Schema) -> str:
  """Convert our schema to duckdb schema."""
  members: list[str] = [
    f'{escape_string_literal(name)}: {escape_string_literal(_duckdb_struct(field))}'
    for name, field in schema.fields.items()
  ]
  return '{' + ', '.join(members) + '}'



"""Tests for the dict source."""

from ..schema import schema
from ..source import SourceSchema
from .dict_source import DictSource


def test_simple_records() -> None:
  items = [{'x': 1, 'y': 'ten'}, {'x': 2, 'y': 'twenty'}]

  source = DictSource(items)
  source.setup()

  source_schema = source.source_schema()
  assert source_schema == SourceSchema(
    fields=schema({'x': 'int64', 'y': 'string'}).fields, num_items=2
  )

  items = list(source.yield_items())
  assert items == [{'x': 1, 'y': 'ten'}, {'x': 2, 'y': 'twenty'}]


def test_inconsistent_records() -> None:
  items = [{'x': 1, 'y': 'ten'}, {'x': 2, 'y': ['twenty']}, {'x': 3, 'y': {'thirty': True}}]

  source = DictSource(items)
  source.setup()

  source_schema = source.source_schema()
  assert source_schema == SourceSchema(
    fields=schema({'x': 'int64', 'y': 'string'}).fields, num_items=3
  )

  items = list(source.yield_items())
  assert items == [
    {'x': 1, 'y': '"ten"'},
    {'x': 2, 'y': '["twenty"]'},
    {'x': 3, 'y': '{"thirty":true}'},
  ]



"""Utilities for inferring dataset formats.

Formats are taken from axlotl: https://github.com/OpenAccess-AI-Collective/axolotl#dataset
"""

import copy
from typing import Any, Callable, ClassVar, Optional, Type, Union

from pydantic import BaseModel, ConfigDict, model_serializer

from .schema import Item, PathTuple, Schema


class DatasetFormatInputSelector(BaseModel):
  """Input lambda selectors that map an item to a string, used for format-specific runtime filters.

  For example, in the ShareGPT format, we want 'human' to only filter conversations
  where conversations.*.from='human'.
  """

  name: str
  selector: Callable[[Item], str]


class DatasetFormat(BaseModel):
  """A dataset format.

  Formats are taken from: https://github.com/OpenAccess-AI-Collective/axolotl#dataset.

  Formats allow us to specify opinionated selectors, like selecting only "human" parts of
  conversations. They also allow us to specify title slots, which are used to display
  titles over certain media fields in the UI.
  """

  # Allow extra fields for dataset formats.
  model_config = ConfigDict(extra='allow')

  name: ClassVar[str]
  data_schema: Optional[Schema] = None

  # Title slots are a mapping of a media path to a path that represents the title to be displayed
  # for that media path. This allows us to put a title over certain media fields in the UI.
  title_slots: Optional[list[tuple[PathTuple, PathTuple]]] = []

  # Input selectors are used for format-specific runtime filters.
  input_selectors: ClassVar[dict[str, DatasetFormatInputSelector]] = {}

  @model_serializer(mode='wrap', when_used='always')
  def serialize_model(self, serializer: Callable[..., dict[str, Any]]) -> dict[str, Any]:
    """Serialize the model to a dictionary."""
    res = serializer(self)
    res['format_name'] = self.name
    return res


def schema_is_compatible_with(dataset_schema: Schema, format_schema: Schema) -> bool:
  """Returns true if all fields of the format schema are in the dataset schema."""
  for path, field in format_schema.all_fields:
    if not dataset_schema.has_field(path):
      return False

    field = dataset_schema.get_field(path)
    if field.dtype != format_schema.get_field(path).dtype:
      return False

  return True


DATASET_FORMAT_REGISTRY: dict[str, Type[DatasetFormat]] = {}


def infer_formats(data_schema: Schema) -> list[type[DatasetFormat]]:
  """Infer the dataset formats for a dataset."""
  formats = []
  for format_cls in DATASET_FORMAT_REGISTRY.values():
    format = format_cls()
    if format.data_schema and schema_is_compatible_with(data_schema, format.data_schema):
      formats.append(format_cls)
  return formats


def register_dataset_format(format_cls: Type[DatasetFormat]) -> None:
  """Register a dataset format in the global registry."""
  if format_cls.name in DATASET_FORMAT_REGISTRY:
    raise ValueError(f'Dataset format "{format_cls.name}" has already been registered!')

  DATASET_FORMAT_REGISTRY[format_cls.name] = format_cls


def get_dataset_format_cls(format_name: str) -> Optional[Type[DatasetFormat]]:
  """Return a registered dataset format class given the name in the registry."""
  return (
    DATASET_FORMAT_REGISTRY.get(format_name) if format_name in DATASET_FORMAT_REGISTRY else None
  )


def resolve_dataset_format(dataset_format: Union[dict, DatasetFormat]) -> DatasetFormat:
  """Resolve a generic format base class to a specific format class."""
  if isinstance(dataset_format, DatasetFormat):
    # The format is already parsed.
    return dataset_format

  format_name = dataset_format.get('format_name')
  if not format_name:
    raise ValueError('"dataset_format" needs to be defined in the json dict.')

  format_cls = get_dataset_format_cls(format_name)
  if not format_cls:
    # Make a metaclass so we get a valid `DatasetFormat` class.
    format_cls = type(f'DatasetFormat_{format_name}', (DatasetFormat,), {'name': format_name})

  dataset_format = copy.deepcopy(dataset_format)
  del dataset_format['dataset_format']
  return format_cls(**dataset_format)


def clear_dataset_format_registry() -> None:
  """Clear the format registry."""
  DATASET_FORMAT_REGISTRY.clear()



"""Project utilities."""
import os
import pathlib
import threading
from typing import Optional, Union

from .config import (
  Config,
  DatasetConfig,
  DatasetSettings,
  EmbeddingConfig,
  SignalConfig,
  get_dataset_config,
)
from .env import env, get_project_dir, set_project_dir
from .utils import log, read_yaml, to_yaml

PROJECT_CONFIG_FILENAME = 'lilac.yml'
PROJECT_CONFIG_LOCK = threading.Lock()


def init(project_dir: Optional[Union[str, pathlib.Path]] = None) -> None:
  """Initializes a project."""
  project_dir = project_dir or get_project_dir()
  if dir_is_project(project_dir):
    raise ValueError(f'{project_dir} is already a project.')

  create_project(project_dir)

  log(f'Successfully initialized project at {project_dir}')


def add_project_dataset_config(
  dataset_config: DatasetConfig,
  project_dir: Optional[Union[str, pathlib.Path]] = None,
  overwrite: bool = False,
) -> None:
  """Add a dataset to the project config.

  Args:
    dataset_config: The dataset configuration to load.
    project_dir: The path to the project directory for where to create the dataset. If not defined,
      uses the project directory from `LILAC_PROJECT_DIR` or [deprecated] `LILAC_DATA_PATH`.
    overwrite: Whether to overwrite the dataset if it already exists.
  """
  project_dir = project_dir or get_project_dir()
  with PROJECT_CONFIG_LOCK:
    config = read_project_config(project_dir)
    existing_dataset_config = get_dataset_config(
      config, dataset_config.namespace, dataset_config.name
    )
    if existing_dataset_config is not None:
      if overwrite:
        config.datasets.remove(existing_dataset_config)
      else:
        raise ValueError(
          f'Dataset {dataset_config.namespace}/{dataset_config.name} has already been added. '
          'Provide `overwrite=True` when loading the data, or delete the old one with:\n\n'
          f'dataset = get_dataset("{dataset_config.namespace}", "{dataset_config.name}")\n'
          'dataset.delete()'
        )

    config.datasets.append(dataset_config)
    write_project_config(project_dir, config)


def delete_project_dataset_config(
  namespace: str, dataset_name: str, project_dir: Optional[Union[str, pathlib.Path]] = None
) -> None:
  """Delete a dataset config in a project."""
  project_dir = project_dir or get_project_dir()

  with PROJECT_CONFIG_LOCK:
    config = read_project_config(project_dir)
    dataset_config = get_dataset_config(config, namespace, dataset_name)
    if dataset_config is None:
      raise ValueError(f'{dataset_config} not found in project config.')

    config.datasets.remove(dataset_config)
    write_project_config(project_dir, config)


def update_project_dataset_settings(
  dataset_namespace: str,
  dataset_name: str,
  settings: DatasetSettings,
  project_dir: Optional[Union[str, pathlib.Path]] = None,
) -> None:
  """Update the settings of a dataset config in a project."""
  project_dir = project_dir or get_project_dir()

  with PROJECT_CONFIG_LOCK:
    config = read_project_config(project_dir)
    dataset_config = get_dataset_config(config, dataset_namespace, dataset_name)
    if dataset_config is None:
      raise ValueError(f'Dataset "{dataset_namespace}/{dataset_name}" not found in project config.')
    dataset_config.settings = settings
    write_project_config(project_dir, config)


def add_project_signal_config(
  dataset_namespace: str,
  dataset_name: str,
  signal_config: SignalConfig,
  project_dir: Optional[Union[str, pathlib.Path]] = None,
) -> None:
  """Add a dataset signal to the project config."""
  project_dir = project_dir or get_project_dir()

  with PROJECT_CONFIG_LOCK:
    config = read_project_config(project_dir)
    dataset_config = get_dataset_config(config, dataset_namespace, dataset_name)
    if dataset_config is None:
      raise ValueError(
        f'Dataset "{dataset_namespace}/{dataset_name}" not found in project config in '
        f'project dir: {project_dir}.'
      )
    if signal_config in dataset_config.signals:
      return
    dataset_config.signals.append(signal_config)
    write_project_config(project_dir, config)


def add_project_embedding_config(
  dataset_namespace: str,
  dataset_name: str,
  embedding_config: EmbeddingConfig,
  project_dir: Optional[Union[str, pathlib.Path]] = None,
) -> None:
  """Add a dataset embedding to the project config."""
  project_dir = project_dir or get_project_dir()

  with PROJECT_CONFIG_LOCK:
    config = read_project_config(project_dir)
    dataset_config = get_dataset_config(config, dataset_namespace, dataset_name)
    if dataset_config is None:
      raise ValueError(f'Dataset "{dataset_namespace}/{dataset_name}" not found in project config.')

    if embedding_config in dataset_config.embeddings:
      return

    dataset_config.embeddings.append(embedding_config)
    write_project_config(project_dir, config)


def delete_project_signal_config(
  dataset_namespace: str,
  dataset_name: str,
  signal_config: SignalConfig,
  project_dir: Optional[Union[str, pathlib.Path]] = None,
) -> None:
  """Delete a dataset signal from the project config."""
  project_dir = project_dir or get_project_dir()

  with PROJECT_CONFIG_LOCK:
    config = read_project_config(project_dir)
    dataset_config = get_dataset_config(config, dataset_namespace, dataset_name)
    if dataset_config is None:
      raise ValueError(f'{dataset_config} not found in project config.')
    dataset_config.signals = [
      s
      for s in dataset_config.signals
      if s.model_dump(exclude_none=True) != signal_config.model_dump(exclude_none=True)
    ]
    write_project_config(project_dir, config)


def project_dir_from_args(project_dir_arg: str) -> str:
  """Returns the project directory from the command line args."""
  project_dir = project_dir_arg
  if project_dir_arg == '':
    project_dir = env('LILAC_PROJECT_DIR', None) or env('LILAC_DATA_PATH', None)
  if not project_dir:
    project_dir = '.'

  return os.path.expanduser(project_dir)


def dir_is_project(project_dir: Union[str, pathlib.Path]) -> bool:
  """Returns whether the directory is a Lilac project."""
  if not os.path.isdir(project_dir):
    return False

  return os.path.exists(os.path.join(project_dir, PROJECT_CONFIG_FILENAME))


def read_project_config(project_dir: Union[str, pathlib.Path]) -> Config:
  """Reads the project config."""
  project_config_filepath = os.path.join(project_dir, PROJECT_CONFIG_FILENAME)
  if not os.path.exists(project_config_filepath):
    create_project(project_dir)

  return Config(**read_yaml(project_config_filepath))


def write_project_config(project_dir: Union[str, pathlib.Path], config: Config) -> None:
  """Writes the project config."""
  with open(os.path.join(project_dir, PROJECT_CONFIG_FILENAME), 'w') as f:
    yaml_config = to_yaml(config.model_dump(exclude_defaults=True, exclude_none=True))
    f.write(
      '# Lilac project config.\n'
      + '# See https://docs.lilacml.com/api_reference/index.html#lilac.Config '
      'for details.\n\n' + yaml_config
    )


def create_project(project_dir: Union[str, pathlib.Path]) -> None:
  """Creates an empty lilac project if it's not already a project."""
  if not dir_is_project(project_dir):
    os.makedirs(project_dir, exist_ok=True)

    write_project_config(project_dir, Config(datasets=[]))


def create_project_and_set_env(project_dir_arg: str) -> None:
  """Creates a Lilac project if it doesn't exist and set the environment variable."""
  project_dir = project_dir_from_args(project_dir_arg)
  create_project(project_dir)
  set_project_dir(project_dir)



"""Tests for dataset.select_rows(sort_by=...)."""

from typing import ClassVar, Iterable, Iterator, Optional, Sequence, cast

import numpy as np
import pytest
from typing_extensions import override

from ..embeddings.vector_store import VectorDBIndex
from ..schema import (
  ROWID,
  Field,
  Item,
  RichData,
  SignalInputType,
  SpanVector,
  VectorKey,
  field,
  lilac_embedding,
  span,
)
from ..signal import (
  TextEmbeddingSignal,
  TextSignal,
  VectorSignal,
  clear_signal_registry,
  register_signal,
)
from .dataset import Column, SortOrder
from .dataset_test_utils import TestDataMaker, enriched_item


class TestSignal(TextSignal):
  name: ClassVar[str] = 'test_signal'

  def fields(self) -> Field:
    return field(fields={'len': 'int32', 'is_all_cap': 'boolean'})

  def compute(self, data: Iterable[RichData]) -> Iterator[Optional[Item]]:
    for text_content in data:
      yield {'len': len(text_content), 'is_all_cap': text_content.isupper()}


class TestPrimitiveSignal(TextSignal):
  name: ClassVar[str] = 'primitive_signal'

  def fields(self) -> Field:
    return field('int32')

  def compute(self, data: Iterable[RichData]) -> Iterator[Optional[Item]]:
    for text_content in data:
      yield len(text_content) + 1


class NestedArraySignal(TextSignal):
  name: ClassVar[str] = 'nested_array'

  def fields(self) -> Field:
    return field(fields=[['int32']])

  def compute(self, data: Iterable[RichData]) -> Iterator[Optional[Item]]:
    for text_content in data:
      yield [[len(text_content) + 1], [len(text_content)]]


@pytest.fixture(scope='module', autouse=True)
def setup_teardown() -> Iterable[None]:
  # Setup.
  register_signal(TestSignal)
  register_signal(TestPrimitiveSignal)
  register_signal(NestedArraySignal)
  register_signal(TopKEmbedding)
  # Unit test runs.
  yield
  # Teardown.
  clear_signal_registry()


def test_sort_by_source_no_alias_no_repeated(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(
    [
      {'erased': True, 'score': 4.1, 'document': {'num_pages': 4, 'header': {'title': 'c'}}},
      {'erased': False, 'score': 3.5, 'document': {'num_pages': 5, 'header': {'title': 'b'}}},
      {'erased': True, 'score': 3.7, 'document': {'num_pages': 3, 'header': {'title': 'a'}}},
    ]
  )

  # Sort by bool.
  result = dataset.select_rows(columns=[ROWID], sort_by=['erased'], sort_order=SortOrder.ASC)
  assert list(result) == [{ROWID: '2'}, {ROWID: '1'}, {ROWID: '3'}]
  result = dataset.select_rows(columns=[ROWID], sort_by=['erased'], sort_order=SortOrder.DESC)
  assert list(result) == [{ROWID: '1'}, {ROWID: '3'}, {ROWID: '2'}]

  # Sort by float.
  result = dataset.select_rows(columns=[ROWID], sort_by=['score'], sort_order=SortOrder.ASC)
  assert list(result) == [{ROWID: '2'}, {ROWID: '3'}, {ROWID: '1'}]
  result = dataset.select_rows(columns=[ROWID], sort_by=['score'], sort_order=SortOrder.DESC)
  assert list(result) == [{ROWID: '1'}, {ROWID: '3'}, {ROWID: '2'}]

  # Sort by nested int.
  result = dataset.select_rows(
    columns=[ROWID], sort_by=['document.num_pages'], sort_order=SortOrder.ASC
  )
  assert list(result) == [{ROWID: '3'}, {ROWID: '1'}, {ROWID: '2'}]
  result = dataset.select_rows(
    columns=[ROWID], sort_by=['document.num_pages'], sort_order=SortOrder.DESC
  )
  assert list(result) == [{ROWID: '2'}, {ROWID: '1'}, {ROWID: '3'}]

  # Sort by double nested string.
  result = dataset.select_rows(
    columns=[ROWID], sort_by=['document.header.title'], sort_order=SortOrder.ASC
  )
  assert list(result) == [{ROWID: '3'}, {ROWID: '2'}, {ROWID: '1'}]
  result = dataset.select_rows(
    columns=[ROWID], sort_by=['document.header.title'], sort_order=SortOrder.DESC
  )
  assert list(result) == [{ROWID: '1'}, {ROWID: '2'}, {ROWID: '3'}]


def test_sort_by_signal_no_alias_no_repeated(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'HEY'}, {'text': 'everyone'}, {'text': 'HI'}])

  dataset.compute_signal(TestSignal(), 'text')

  # Sort by `signal.len`.
  result = dataset.select_rows(
    columns=[ROWID], sort_by=['text.test_signal.len'], sort_order=SortOrder.ASC
  )
  assert list(result) == [{ROWID: '3'}, {ROWID: '1'}, {ROWID: '2'}]
  result = dataset.select_rows(
    columns=[ROWID], sort_by=['text.test_signal.len'], sort_order=SortOrder.DESC
  )
  assert list(result) == [{ROWID: '2'}, {ROWID: '1'}, {ROWID: '3'}]

  # Sort by `signal.is_all_cap`.
  result = dataset.select_rows(
    columns=[ROWID], sort_by=['text.test_signal.is_all_cap'], sort_order=SortOrder.ASC
  )
  assert list(result) == [{ROWID: '2'}, {ROWID: '1'}, {ROWID: '3'}]
  result = dataset.select_rows(
    columns=[ROWID], sort_by=['text.test_signal.is_all_cap'], sort_order=SortOrder.DESC
  )
  assert list(result) == [{ROWID: '1'}, {ROWID: '3'}, {ROWID: '2'}]


def test_sort_by_signal_alias_no_repeated(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'HEY'}, {'text': 'everyone'}, {'text': 'HI'}])

  dataset.compute_signal(TestSignal(), 'text')

  # Sort by `signal.len`.
  signal_alias = Column('text.test_signal', alias='signal')
  result = dataset.select_rows(
    columns=[signal_alias], sort_by=['signal.len'], sort_order=SortOrder.ASC
  )
  assert list(result) == [
    {'signal': {'len': 2, 'is_all_cap': True}},
    {'signal': {'len': 3, 'is_all_cap': True}},
    {'signal': {'len': 8, 'is_all_cap': False}},
  ]
  result = dataset.select_rows(
    columns=[signal_alias], sort_by=['signal.len'], sort_order=SortOrder.DESC
  )
  assert list(result) == [
    {'signal': {'len': 8, 'is_all_cap': False}},
    {'signal': {'len': 3, 'is_all_cap': True}},
    {'signal': {'len': 2, 'is_all_cap': True}},
  ]


def test_sort_by_enriched_alias_no_repeated(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'HEY'}, {'text': 'everyone'}, {'text': 'HI'}])

  dataset.compute_signal(TestSignal(), 'text')

  # Sort by `document.test_signal.is_all_cap` where 'document' is an alias to 'text'.
  text_alias = Column('text', alias='document')
  result = dataset.select_rows(
    columns=[text_alias],
    sort_by=['document.test_signal.is_all_cap'],
    sort_order=SortOrder.ASC,
    combine_columns=True,
  )
  assert list(result) == [
    {'text': enriched_item('everyone', {'test_signal': {'len': 8, 'is_all_cap': False}})},
    {'text': enriched_item('HEY', {'test_signal': {'len': 3, 'is_all_cap': True}})},
    {'text': enriched_item('HI', {'test_signal': {'len': 2, 'is_all_cap': True}})},
  ]

  result = dataset.select_rows(
    columns=[text_alias],
    sort_by=['document.test_signal.is_all_cap'],
    sort_order=SortOrder.DESC,
    combine_columns=True,
  )
  # Aliases are ignored when combining columns.
  assert list(result) == [
    {'text': enriched_item('HEY', {'test_signal': {'len': 3, 'is_all_cap': True}})},
    {'text': enriched_item('HI', {'test_signal': {'len': 2, 'is_all_cap': True}})},
    {'text': enriched_item('everyone', {'test_signal': {'len': 8, 'is_all_cap': False}})},
  ]


def test_sort_by_udf_alias_no_repeated(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'HEY'}, {'text': 'everyone'}, {'text': 'HI'}])

  # Equivalent to: SELECT `TestSignal(text) AS udf`.
  text_udf = Column('text', signal_udf=TestSignal(), alias='udf')
  # Sort by `udf.len`, where `udf` is an alias to `TestSignal(text)`.
  result = dataset.select_rows(['*', text_udf], sort_by=['udf.len'], sort_order=SortOrder.ASC)
  assert list(result) == [
    {'text': 'HI', 'udf': {'len': 2, 'is_all_cap': True}},
    {'text': 'HEY', 'udf': {'len': 3, 'is_all_cap': True}},
    {'text': 'everyone', 'udf': {'len': 8, 'is_all_cap': False}},
  ]


def test_sort_by_udf_no_alias_no_repeated(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'HEY'}, {'text': 'everyone'}, {'text': 'HI'}])

  text_udf = Column('text', signal_udf=TestSignal())
  # Sort by `text.test_signal.len`, produced by executing the udf `TestSignal(text)`.
  result = dataset.select_rows(
    ['*', text_udf],
    sort_by=[('text', 'test_signal', 'len')],
    sort_order=SortOrder.ASC,
    combine_columns=True,
  )
  assert list(result) == [
    {'text': enriched_item('HI', {'test_signal': {'len': 2, 'is_all_cap': True}})},
    {'text': enriched_item('HEY', {'test_signal': {'len': 3, 'is_all_cap': True}})},
    {'text': enriched_item('everyone', {'test_signal': {'len': 8, 'is_all_cap': False}})},
  ]

  # Sort descending.
  result = dataset.select_rows(
    ['*', text_udf],
    sort_by=[('text', 'test_signal', 'len')],
    sort_order=SortOrder.DESC,
    combine_columns=True,
  )
  assert list(result) == [
    {'text': enriched_item('everyone', {'test_signal': {'len': 8, 'is_all_cap': False}})},
    {'text': enriched_item('HEY', {'test_signal': {'len': 3, 'is_all_cap': True}})},
    {'text': enriched_item('HI', {'test_signal': {'len': 2, 'is_all_cap': True}})},
  ]


def test_sort_by_primitive_udf_alias_no_repeated(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'HEY'}, {'text': 'everyone'}, {'text': 'HI'}])

  # Equivalent to: SELECT `TestPrimitiveSignal(text) AS udf`.
  text_udf = Column('text', signal_udf=TestPrimitiveSignal(), alias='udf')
  # Sort by the primitive value returned by the udf.
  result = dataset.select_rows(['*', text_udf], sort_by=['udf'], sort_order=SortOrder.ASC)
  assert list(result) == [
    {'text': 'HI', 'udf': 3},
    {'text': 'HEY', 'udf': 4},
    {'text': 'everyone', 'udf': 9},
  ]


def test_sort_by_source_non_leaf_errors(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'vals': [7, 1]}, {'vals': [3, 4]}, {'vals': [9, 0]}])

  # Sort by repeated.
  with pytest.raises(ValueError, match='Unable to sort by path'):
    dataset.select_rows(columns=[ROWID], sort_by=['vals'], sort_order=SortOrder.ASC)


def test_sort_by_source_no_alias_repeated(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(
    [
      {'vals': [[{'score': 7}, {'score': 1}], [{'score': 1}, {'score': 7}]]},
      {'vals': [[{'score': 3}, {'score': 4}]]},
      {'vals': [[{'score': 9}, {'score': 0}]]},
    ]
  )

  # Sort by repeated 'vals'.
  result = dataset.select_rows(
    columns=['vals'], sort_by=['vals.*.*.score'], sort_order=SortOrder.ASC
  )
  assert list(result) == [
    {'vals': [[{'score': 9}, {'score': 0}]]},
    {'vals': [[{'score': 7}, {'score': 1}], [{'score': 1}, {'score': 7}]]},
    {'vals': [[{'score': 3}, {'score': 4}]]},
  ]

  result = dataset.select_rows(
    columns=['vals'], sort_by=['vals.*.*.score'], sort_order=SortOrder.DESC
  )
  assert list(result) == [
    {'vals': [[{'score': 9}, {'score': 0}]]},
    {'vals': [[{'score': 7}, {'score': 1}], [{'score': 1}, {'score': 7}]]},
    {'vals': [[{'score': 3}, {'score': 4}]]},
  ]


def test_sort_by_source_alias_repeated(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'vals': [[7, 1], [1, 7]]}, {'vals': [[3], [11]]}, {'vals': [[9, 0]]}])

  # Sort by repeated 'vals'.
  result = dataset.select_rows(
    columns=[Column('vals', alias='scores')], sort_by=['scores.*.*'], sort_order=SortOrder.ASC
  )
  assert list(result) == [
    {'scores': [[9, 0]]},
    {'scores': [[7, 1], [1, 7]]},
    {'scores': [[3], [11]]},
  ]

  result = dataset.select_rows(
    columns=[Column('vals', alias='scores')], sort_by=['scores.*.*'], sort_order=SortOrder.DESC
  )
  assert list(result) == [
    {'scores': [[3], [11]]},
    {'scores': [[9, 0]]},
    {'scores': [[7, 1], [1, 7]]},
  ]


def test_sort_by_udf_alias_repeated(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'HEY'}, {'text': 'everyone'}, {'text': 'HI'}])

  # Equivalent to: SELECT `NestedArraySignal(text) AS udf`.
  text_udf = Column('text', signal_udf=NestedArraySignal(), alias='udf')
  # Sort by `udf.*.*`, where `udf` is an alias to `NestedArraySignal(text)`.
  result = dataset.select_rows(['*', text_udf], sort_by=['udf.*.*'], sort_order=SortOrder.ASC)
  assert list(result) == [
    {'text': 'HI', 'udf': [[3], [2]]},
    {'text': 'HEY', 'udf': [[4], [3]]},
    {'text': 'everyone', 'udf': [[9], [8]]},
  ]
  result = dataset.select_rows(['*', text_udf], sort_by=['udf.*.*'], sort_order=SortOrder.DESC)
  assert list(result) == [
    {'text': 'everyone', 'udf': [[9], [8]]},
    {'text': 'HEY', 'udf': [[4], [3]]},
    {'text': 'HI', 'udf': [[3], [2]]},
  ]


def test_sort_by_complex_signal_udf_alias_called_on_repeated(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(
    [
      {'texts': [{'text': 'eardrop'}, {'text': 'I'}]},
      {'texts': [{'text': 'hey'}, {'text': 'CARS'}]},
      {'texts': [{'text': 'everyone'}, {'text': ''}]},
    ]
  )

  # Equivalent to: SELECT `TestSignal(texts.*.text) AS udf`.
  texts_udf = Column('texts.*.text', signal_udf=TestSignal(), alias='udf')
  # Sort by `udf.len`, where `udf` is an alias to `TestSignal(texts.*.text)`.
  result = dataset.select_rows(
    ['*', texts_udf], sort_by=['udf.len'], sort_order=SortOrder.ASC, combine_columns=True
  )
  assert list(result) == [
    {
      'texts': [
        {'text': enriched_item('everyone', {'test_signal': {'len': 8, 'is_all_cap': False}})},
        {'text': enriched_item('', {'test_signal': {'len': 0, 'is_all_cap': False}})},
      ]
    },
    {
      'texts': [
        {'text': enriched_item('eardrop', {'test_signal': {'len': 7, 'is_all_cap': False}})},
        {'text': enriched_item('I', {'test_signal': {'len': 1, 'is_all_cap': True}})},
      ]
    },
    {
      'texts': [
        {'text': enriched_item('hey', {'test_signal': {'len': 3, 'is_all_cap': False}})},
        {'text': enriched_item('CARS', {'test_signal': {'len': 4, 'is_all_cap': True}})},
      ]
    },
  ]


def test_sort_by_primitive_signal_udf_alias_called_on_repeated(
  make_test_data: TestDataMaker
) -> None:
  dataset = make_test_data(
    [
      {'texts': [{'text': 'eardrop'}, {'text': 'I'}]},
      {'texts': [{'text': 'hey'}, {'text': 'CARS'}]},
      {'texts': [{'text': 'everyone'}, {'text': ''}]},
    ]
  )

  # Equivalent to: SELECT `TestPrimitiveSignal(texts.*.text) AS udf`.
  texts_udf = Column('texts.*.text', signal_udf=TestPrimitiveSignal(), alias='udf')
  # Sort by `udf`, where `udf` is an alias to `TestPrimitiveSignal(texts.*.text)`.
  result = dataset.select_rows(
    ['*', texts_udf], sort_by=['udf'], sort_order=SortOrder.ASC, combine_columns=True
  )
  assert list(result) == [
    {
      'texts': [
        {'text': enriched_item('everyone', {'primitive_signal': 9})},
        {'text': enriched_item('', {'primitive_signal': 1})},
      ]
    },
    {
      'texts': [
        {'text': enriched_item('eardrop', {'primitive_signal': 8})},
        {'text': enriched_item('I', {'primitive_signal': 2})},
      ]
    },
    {
      'texts': [
        {'text': enriched_item('hey', {'primitive_signal': 4})},
        {'text': enriched_item('CARS', {'primitive_signal': 5})},
      ]
    },
  ]
  result = dataset.select_rows(
    ['*', texts_udf], sort_by=['udf'], sort_order=SortOrder.DESC, combine_columns=True
  )
  assert list(result) == [
    {
      'texts': [
        {'text': enriched_item('everyone', {'primitive_signal': 9})},
        {'text': enriched_item('', {'primitive_signal': 1})},
      ]
    },
    {
      'texts': [
        {'text': enriched_item('eardrop', {'primitive_signal': 8})},
        {'text': enriched_item('I', {'primitive_signal': 2})},
      ]
    },
    {
      'texts': [
        {'text': enriched_item('hey', {'primitive_signal': 4})},
        {'text': enriched_item('CARS', {'primitive_signal': 5})},
      ]
    },
  ]


class TopKEmbedding(TextEmbeddingSignal):
  """A test embed function."""

  name: ClassVar[str] = 'topk_embedding'

  def compute(self, data: Iterable[RichData]) -> Iterator[Item]:
    """Call the embedding function."""
    for example in data:
      example = cast(str, example)
      emb_spans: list[Item] = []
      for i, score in enumerate(example.split('_')):
        start, end = i * 2, i * 2 + 1
        vector = np.array([int(score)])
        emb_spans.append(lilac_embedding(start, end, vector))
      yield emb_spans


class TopKSignal(VectorSignal):
  """Compute scores along a given concept for documents."""

  name: ClassVar[str] = 'topk_signal'
  input_type: ClassVar[SignalInputType] = SignalInputType.TEXT

  _query = np.array([1])

  def fields(self) -> Field:
    return field(fields=[field('string_span', {'score': 'float32'})])

  @override
  def vector_compute(
    self, all_vector_spans: Iterable[list[SpanVector]]
  ) -> Iterator[Optional[Item]]:
    for vector_spans in all_vector_spans:
      embeddings = np.array([vector_span['vector'] for vector_span in vector_spans])
      scores = embeddings.dot(self._query).reshape(-1)
      res: Item = []
      for vector_span, score in zip(vector_spans, scores):
        start, end = vector_span['span']
        res.append(span(start, end, {'score': score}))
      yield res

  @override
  def vector_compute_topk(
    self, topk: int, vector_index: VectorDBIndex, rowids: Optional[Iterable[str]] = None
  ) -> Sequence[tuple[VectorKey, Optional[Item]]]:
    return vector_index.topk(self._query, topk, rowids)


def test_sort_by_topk_embedding_udf(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'scores': '8_1'}, {'scores': '3_5'}, {'scores': '9_7'}])

  dataset.compute_signal(TopKEmbedding(), 'scores')

  # Equivalent to: SELECT `TopKSignal(scores, embedding='...') AS udf`.
  signal = TopKSignal(embedding='topk_embedding')
  text_udf = Column('scores', signal_udf=signal, alias='udf')
  # Sort by `udf`, where `udf` is an alias to `TopKSignal(scores, embedding='...')`.
  result = dataset.select_rows(
    ['*', text_udf], sort_by=['udf'], sort_order=SortOrder.DESC, limit=2, combine_columns=True
  )
  assert list(result) == [
    {
      'scores': enriched_item(
        '9_7', {signal.key(): [span(0, 1, {'score': 9.0}), span(2, 3, {'score': 7.0})]}
      )
    },
    {
      'scores': enriched_item(
        '8_1', {signal.key(): [span(0, 1, {'score': 8.0}), span(2, 3, {'score': 1.0})]}
      )
    },
  ]

  # Same but set limit to 3.
  result = dataset.select_rows(
    ['*', text_udf], sort_by=['udf'], sort_order=SortOrder.DESC, limit=3, combine_columns=True
  )
  assert list(result) == [
    {
      'scores': enriched_item(
        '9_7', {signal.key(): [span(0, 1, {'score': 9.0}), span(2, 3, {'score': 7.0})]}
      )
    },
    {
      'scores': enriched_item(
        '8_1', {signal.key(): [span(0, 1, {'score': 8.0}), span(2, 3, {'score': 1.0})]}
      )
    },
    {
      'scores': enriched_item(
        '3_5', {signal.key(): [span(0, 1, {'score': 3.0}), span(2, 3, {'score': 5.0})]}
      )
    },
  ]


def test_sort_by_topk_udf_with_filter(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(
    [
      {'scores': '8_1', 'active': True},
      {'scores': '3_5', 'active': True},
      {'scores': '9_7', 'active': False},
    ]
  )

  dataset.compute_signal(TopKEmbedding(), 'scores')

  # Equivalent to: SELECT `TopKSignal(scores, embedding='...') AS udf`.
  signal = TopKSignal(embedding='topk_embedding')
  text_udf = Column('scores', signal_udf=signal, alias='udf')
  # Sort by `udf`, where `udf` is an alias to `TopKSignal(scores, embedding='...')`.
  result = dataset.select_rows(
    ['*', text_udf],
    sort_by=['udf'],
    filters=[('active', 'equals', True)],
    sort_order=SortOrder.DESC,
    limit=2,
    combine_columns=True,
  )
  # We make sure that '3' is not in the result, because it is not active, even though it has the
  # highest topk score.
  assert list(result) == [
    {
      'active': True,
      'scores': enriched_item(
        '8_1', {signal.key(): [span(0, 1, {'score': 8.0}), span(2, 3, {'score': 1.0})]}
      ),
    },
    {
      'active': True,
      'scores': enriched_item(
        '3_5', {signal.key(): [span(0, 1, {'score': 3.0}), span(2, 3, {'score': 5.0})]}
      ),
    },
  ]



"""Configurations for a dataset run."""

import json
import pathlib
from typing import Literal, Optional, Union

import yaml
from pydantic import (
  BaseModel,
  ConfigDict,
  SerializeAsAny,
  ValidationError,
  ValidationInfo,
  field_serializer,
  field_validator,
)
from pydantic import Field as PydanticField

from .embeddings.embedding import EMBEDDING_SORT_PRIORITIES
from .schema import Path, PathTuple, normalize_path
from .signal import Signal, TextEmbeddingSignal, get_signal_by_type, resolve_signal
from .source import Source, resolve_source

DEFAULT_EMBEDDING = EMBEDDING_SORT_PRIORITIES[0]

OLD_CONFIG_FILENAME = 'config.yml'


def _serializable_path(path: PathTuple) -> Union[str, list]:
  if len(path) == 1:
    return path[0]
  return list(path)


class SignalConfig(BaseModel):
  """Configures a signal on a source path."""

  path: PathTuple
  signal: SerializeAsAny[Signal]
  model_config = ConfigDict(extra='forbid')

  @field_validator('path', mode='before')
  @classmethod
  def parse_path(cls, path: Path) -> PathTuple:
    """Parse a path."""
    return normalize_path(path)

  @field_serializer('path')
  def serialize_path(self, path: PathTuple) -> Union[str, list[str]]:
    """Serialize a path."""
    return _serializable_path(path)

  @field_validator('signal', mode='before')
  @classmethod
  def parse_signal(cls, signal: dict) -> Signal:
    """Parse a signal to its specific subclass instance."""
    return resolve_signal(signal)


class EmbeddingConfig(BaseModel):
  """Configures an embedding on a source path."""

  path: PathTuple
  embedding: str
  model_config = ConfigDict(extra='forbid')

  @field_validator('path', mode='before')
  @classmethod
  def parse_path(cls, path: Path) -> PathTuple:
    """Parse a path."""
    return normalize_path(path)

  @field_serializer('path')
  def serialize_path(self, path: PathTuple) -> Union[str, list[str]]:
    """Serialize a path."""
    return _serializable_path(path)

  @field_validator('embedding', mode='before')
  @classmethod
  def validate_embedding(cls, embedding: str) -> str:
    """Validate the embedding is registered."""
    get_signal_by_type(embedding, TextEmbeddingSignal)
    return embedding


class DatasetUISettings(BaseModel):
  """The UI persistent settings for a dataset."""

  media_paths: list[PathTuple] = []
  markdown_paths: list[PathTuple] = []
  model_config = ConfigDict(extra='forbid')
  view_type: Literal['scroll', 'single_item'] = 'single_item'
  # Maps a label to a keycode.
  label_to_keycode: dict[str, str] = {}

  @field_validator('media_paths', mode='before')
  @classmethod
  def parse_media_paths(cls, media_paths: list) -> list:
    """Parse a path, ensuring it is a tuple."""
    return [normalize_path(path) for path in media_paths]

  @field_validator('markdown_paths', mode='before')
  @classmethod
  def parse_markdown_paths(cls, markdown_paths: list) -> list:
    """Parse a path, ensuring it is a tuple."""
    return [normalize_path(path) for path in markdown_paths]

  @field_serializer('media_paths')
  def serialize_media_paths(self, media_paths: list[PathTuple]) -> list[Union[str, list[str]]]:
    """Serialize media paths."""
    return [_serializable_path(path) for path in media_paths]

  @field_serializer('markdown_paths')
  def serialize_markdown_paths(
    self, markdown_paths: list[PathTuple]
  ) -> list[Union[str, list[str]]]:
    """Serialize markdown paths."""
    return [_serializable_path(path) for path in markdown_paths]


class DatasetSettings(BaseModel):
  """The persistent settings for a dataset."""

  ui: DatasetUISettings = DatasetUISettings()
  preferred_embedding: Optional[str] = DEFAULT_EMBEDDING
  model_config = ConfigDict(extra='forbid')
  tags: Optional[list[str]] = PydanticField(
    description='A list of tags for the dataset to organize in the UI.', default=[]
  )


class ClusterInputSelectorConfig(BaseModel):
  """Configures a format selector for a cluster input."""

  format: str
  selector: str


class ClusterConfig(BaseModel):
  """Configuration for computing clusters for a dataset.

  This is not used in any Lilac-internal database configuration settings. Just some sugar for
  specifying which clusters to compute for our demo website.
  """

  dataset_namespace: str
  dataset_name: str
  input_path: Optional[PathTuple] = None
  input_selector: Optional[ClusterInputSelectorConfig] = None
  output_path: Optional[PathTuple] = None
  min_cluster_size: int = 5
  remote: bool = (
    True  # Whether to compute the cluster via remote service. Toggle to false for unit testing.
  )

  @field_validator('input_selector')
  def check_inputs(
    cls, input_selector: Optional[ClusterInputSelectorConfig], values: ValidationInfo
  ) -> Optional[ClusterInputSelectorConfig]:
    """Check that either `input_path` or `input_selector` is defined."""
    if 'input_path' not in values.data and not input_selector:
      raise ValueError('either `input_path` or `input_selector` is required')
    return input_selector


class DatasetConfig(BaseModel):
  """Configures a dataset with a source and transformations."""

  namespace: str = PydanticField(description='The namespace of the dataset.')
  name: str = PydanticField(description='The name of the dataset.')

  # Deprecated.
  tags: Optional[list[str]] = PydanticField(
    description='[Deprecated] This field is *deprecated* in favor of DatasetSettings.tags and '
    'will be removed in a later release.',
    default=[],
  )

  # The source configuration.
  source: SerializeAsAny[Source] = PydanticField(
    description='The source configuration. This config determines where data is loaded '
    'from for the dataset.'
  )

  # Model configuration: embeddings and signals on paths.
  embeddings: list[EmbeddingConfig] = PydanticField(
    description='The embedding configs for the dataset.', default=[]
  )

  # When defined, uses this list of signals instead of running all signals.
  signals: list[SignalConfig] = PydanticField(
    description='The signal configs for the dataset', default=[]
  )

  # Dataset settings, default embeddings and UI settings like media paths.
  settings: Optional[DatasetSettings] = PydanticField(description='Dataset settings.', default=None)
  model_config = ConfigDict(extra='forbid')

  @field_validator('source', mode='before')
  @classmethod
  def parse_source(cls, source: dict) -> Source:
    """Parse a source to its specific subclass instance."""
    return resolve_source(source)


class Config(BaseModel):
  """Configures a set of datasets for a lilac instance."""

  datasets: list[DatasetConfig] = PydanticField(
    description='The configurations for the datasets in the project.', default=[]
  )

  # When defined, uses this list of signals to run over every dataset, over all media paths, unless
  # signals is overridden by a specific dataset.
  signals: list[Signal] = PydanticField(
    description='The signals to run for every dataset.', default=[]
  )

  # A list of embeddings to compute the model caches for, for all concepts.
  concept_model_cache_embeddings: list[str] = PydanticField(
    description='The set of embeddings to compute model caches for for every concept.', default=[]
  )

  clusters: list[ClusterConfig] = PydanticField(
    description='The set of datasets and paths to compute clusters for.', default=[]
  )
  model_config = ConfigDict(extra='forbid')

  @field_validator('signals', mode='before')
  @classmethod
  def parse_signal(cls, signals: list[dict]) -> list[Signal]:
    """Parse alist of signals to their specific subclass instances."""
    return [resolve_signal(signal) for signal in signals]


def get_dataset_config(
  config: Config, dataset_namespace: str, dataset_name: str
) -> Optional[DatasetConfig]:
  """Returns the dataset config."""
  for dataset_config in config.datasets:
    if dataset_config.namespace == dataset_namespace and dataset_config.name == dataset_name:
      return dataset_config

  return None


def read_config(config_path: Union[str, pathlib.Path]) -> Config:
  """Reads a config file.

  The config file can either be a `Config` or a `DatasetConfig`.

  The result is always a `Config` object. If the input is a `DatasetConfig`, the config will just
  contain a single dataset.
  """
  config_ext = pathlib.Path(config_path).suffix
  if config_ext in ['.yml', '.yaml']:
    with open(config_path, 'r') as f:
      config_dict = yaml.safe_load(f)
  elif config_ext in ['.json']:
    with open(config_path, 'r') as f:
      config_dict = json.load(f)
  else:
    raise ValueError(f'Unsupported config file extension: {config_ext}')

  config: Optional[Config] = None
  is_config = True
  try:
    config = Config(**config_dict)
  except ValidationError:
    is_config = False

  if not is_config:
    try:
      dataset_config = DatasetConfig(**config_dict)
      config = Config(datasets=[dataset_config])
    except ValidationError as error:
      raise ValueError('Config is not a valid `Config` or `DatasetConfig`') from error
  assert config is not None

  return config



"""A signal to compute a score along a concept."""
import os
from typing import ClassVar, Iterable, Iterator, Optional

import numpy as np
from typing_extensions import override

from ..auth import UserInfo
from ..batch_utils import flat_batched_compute
from ..concepts.concept import DRAFT_MAIN, ConceptModel
from ..concepts.db_concept import DISK_CONCEPT_MODEL_DB, ConceptModelDB
from ..embeddings.embedding import get_embed_fn
from ..embeddings.vector_store import VectorDBIndex
from ..schema import Field, Item, PathKey, RichData, SignalInputType, SpanVector, field, span
from ..signal import VectorSignal


class ConceptSignal(VectorSignal):
  """Compute scores along a given concept for documents."""

  name: ClassVar[str] = 'concept_score'
  input_type: ClassVar[SignalInputType] = SignalInputType.TEXT

  display_name: ClassVar[str] = 'Concept'

  namespace: str
  concept_name: str
  # This is set during setup and used in the UI to show the concept version on hover.
  version: Optional[int] = None

  # The draft version of the concept to use. If not provided, the latest version is used.
  draft: str = DRAFT_MAIN

  _concept_model_db: ConceptModelDB = DISK_CONCEPT_MODEL_DB
  _user: Optional[UserInfo] = None

  @override
  def fields(self) -> Field:
    return field(
      fields=[
        field(
          dtype='string_span',
          fields={
            'score': field(
              'float32', bins=[('Not in concept', None, 0.5), ('In concept', 0.5, None)]
            )
          },
        )
      ]
    )

  def set_user(self, user: Optional[UserInfo]) -> None:
    """Set the user for this signal."""
    self._user = user

  def _get_concept_model(self) -> ConceptModel:
    return self._concept_model_db.sync(
      self.namespace, self.concept_name, self.embedding, self._user, create=True
    )

  def _score_span_vectors(
    self, span_vectors: Iterable[Iterable[SpanVector]]
  ) -> Iterator[Optional[Item]]:
    concept_model = self._get_concept_model()

    return flat_batched_compute(
      span_vectors,
      f=lambda vectors: self._compute_span_vector_batch(vectors, concept_model),
      batch_size=concept_model.batch_size,
    )

  def _compute_span_vector_batch(
    self, span_vectors: Iterable[SpanVector], concept_model: ConceptModel
  ) -> list[Item]:
    vectors = [sv['vector'] for sv in span_vectors]
    spans = [sv['span'] for sv in span_vectors]
    scores = concept_model.score_embeddings(self.draft, np.array(vectors)).tolist()
    return [span(start, end, {'score': score}) for score, (start, end) in zip(scores, spans)]

  @override
  def setup(self) -> None:
    concept_model = self._get_concept_model()
    self.version = concept_model.version

  @override
  def compute(self, examples: Iterable[RichData]) -> Iterator[Optional[Item]]:
    """Get the scores for the provided examples."""
    embed_fn = get_embed_fn(self.embedding, split=True)
    span_vectors = embed_fn(examples)
    return self._score_span_vectors(span_vectors)

  @override
  def vector_compute(self, span_vectors: Iterable[list[SpanVector]]) -> Iterator[Optional[Item]]:
    return self._score_span_vectors(span_vectors)

  @override
  def vector_compute_topk(
    self, topk: int, vector_index: VectorDBIndex, rowids: Optional[Iterable[str]] = None
  ) -> list[tuple[PathKey, Optional[Item]]]:
    concept_model = self._get_concept_model()
    query: np.ndarray = concept_model.coef(self.draft).astype(np.float32)
    query /= np.linalg.norm(query)
    topk_keys = [key for key, _ in vector_index.topk(query, topk, rowids)]
    return list(zip(topk_keys, self.vector_compute(vector_index.get(topk_keys))))

  @override
  def key(self, is_computed_signal: Optional[bool] = False) -> str:
    path = [self.namespace, self.concept_name, self.embedding]
    if not is_computed_signal:
      path.append('preview')
    return os.path.join(*path)



"""Jina embeddings. Open-source, designed to run on device, with 8K context."""
import gc
from typing import TYPE_CHECKING, Any, ClassVar, Iterator, Optional, cast

import modal

from ..batch_utils import compress_docs
from ..embeddings.embedding import chunked_compute_embedding
from ..tasks import TaskExecutionType
from .transformer_utils import setup_model_device

if TYPE_CHECKING:
  from transformers import AutoModel

import numpy as np
from numpy.linalg import norm
from typing_extensions import override

from ..schema import Item, lilac_embedding
from ..signal import TextEmbeddingSignal

# See readme in https://huggingface.co/jinaai/jina-embeddings-v2-small-en
_SIZE_TO_MODEL: dict[str, str] = {
  'small': 'jina-embeddings-v2-small-en',
  'base': 'jina-embeddings-v2-base-en',
}

# Anything larger than 1 slows down the computation because a single long document will cause
# padding to be added to all other documents in the batch.
JINA_BATCH_SIZE = 1
JINA_CONTEXT_SIZE = 8192


class JinaV2Small(TextEmbeddingSignal):
  """Jina V2 Embeddings with 8K context.

  Each document is truncated to 8K characters, and the embeddings are computed on the truncated
  document.
  """

  name: ClassVar[str] = 'jina-v2-small'
  display_name: ClassVar[str] = 'Jina V2 (small)'
  local_batch_size: ClassVar[int] = JINA_BATCH_SIZE
  local_parallelism: ClassVar[int] = 1
  local_strategy: ClassVar[TaskExecutionType] = 'threads'
  runs_remote: ClassVar[bool] = True

  _size = 'small'
  _model: Optional['AutoModel'] = None

  @override
  def setup(self) -> None:
    try:
      from transformers import AutoModel
    except ImportError:
      raise ImportError(
        'Could not import the `transformers` python package. '
        'Please install it with `pip install transformers`.'
      )
    # trust_remote_code is needed to use the encode method.
    model_name = f'jinaai/{_SIZE_TO_MODEL[self._size]}'
    self._model = setup_model_device(
      AutoModel.from_pretrained(model_name, trust_remote_code=True), model_name
    )

  @override
  def teardown(self) -> None:
    if self._model is None:
      return
    self._model.cpu()
    del self._model
    try:
      import torch

      torch.cuda.empty_cache()
    except ImportError:
      pass
    gc.collect()

  @override
  def compute(self, docs: list[str]) -> list[Item]:
    """Call the embedding function."""
    if self._model is None:
      raise ValueError('The signal is not initialized. Call setup() first.')

    # SentenceTransformers can take arbitrarily large batches.
    def _embed_fn(docs: list[str]) -> list[np.ndarray]:
      trimmed_docs = [doc[:JINA_CONTEXT_SIZE] for doc in docs]
      vectors = cast(Any, self._model).encode(trimmed_docs)
      embeddings = []
      for vector in vectors:
        vector = np.array(vector)
        vector /= norm(vector)
        embeddings.append(vector)
      return embeddings

    return chunked_compute_embedding(
      _embed_fn,
      docs,
      self.local_batch_size,
    )

  @override
  def compute_remote(self, docs: Iterator[str]) -> Iterator[Item]:
    trimmed_docs: list[str] = []
    doc_lengths: list[int] = []
    for doc in docs:
      trimmed_docs.append(doc[:JINA_CONTEXT_SIZE])
      doc_lengths.append(len(doc))
    gzipped_docs = compress_docs(trimmed_docs)

    del trimmed_docs, docs
    gc.collect()

    index = 0
    jina_batch = modal.Function.lookup('jina-batch', 'embed')
    for batch in jina_batch.remote_gen({'gzipped_docs': gzipped_docs}):
      batch /= norm(batch, axis=1, keepdims=True)
      for vector in batch:
        yield [lilac_embedding(start=0, end=doc_lengths[index], embedding=vector)]
        index += 1


class JinaV2Base(JinaV2Small):
  """Jina V2 Embeddings with 8K context.

  Each document is truncated to 8K characters, and the embeddings are computed on the truncated
  document.
  """

  name: ClassVar[str] = 'jina-v2-base'
  display_name: ClassVar[str] = 'Jina V2 (base)'

  runs_remote: ClassVar[bool] = False
  _size = 'base'



"""Tests for dataset.compute_signal()."""

import re
from typing import ClassVar, Iterable, Iterator, Optional, Union, cast

import numpy as np
import pytest
from pytest_mock import MockerFixture
from typing_extensions import override

from ..concepts.concept import ExampleIn
from ..concepts.db_concept import ConceptUpdate, DiskConceptDB
from ..schema import (
  EMBEDDING_KEY,
  PATH_WILDCARD,
  Field,
  Item,
  RichData,
  SignalInputType,
  field,
  lilac_embedding,
  schema,
  span,
)
from ..signal import TextEmbeddingSignal, TextSignal, clear_signal_registry, register_signal
from ..signals.concept_scorer import ConceptSignal
from ..source import clear_source_registry, register_source
from . import dataset_utils as dataset_utils_module
from .dataset import Column, DatasetManifest, GroupsSortBy, SortOrder
from .dataset_test_utils import (
  TEST_DATASET_NAME,
  TEST_NAMESPACE,
  TestDataMaker,
  TestSource,
  enriched_item,
)

SIMPLE_ITEMS: list[Item] = [
  {'str': 'a', 'int': 1, 'bool': False, 'float': 3.0},
  {'str': 'b', 'int': 2, 'bool': True, 'float': 2.0},
  {'str': 'b', 'int': 2, 'bool': True, 'float': 1.0},
]


class TestSparseSignal(TextSignal):
  name: ClassVar[str] = 'test_sparse_signal'

  @override
  def fields(self) -> Field:
    return field('int32')

  @override
  def compute(self, data: Iterable[RichData]) -> Iterator[Optional[Item]]:
    for text in data:
      if text == 'hello':
        # Skip this input.
        yield None
      else:
        yield len(text)


class TestSparseRichSignal(TextSignal):
  """Find personally identifiable information (emails, phone numbers, etc)."""

  name: ClassVar[str] = 'test_sparse_rich_signal'

  @override
  def fields(self) -> Field:
    return field(fields={'emails': ['string']})

  @override
  def compute(self, data: Iterable[RichData]) -> Iterator[Optional[Item]]:
    for text in data:
      if text == 'hello':
        # Skip this input.
        yield None
      else:
        yield {'emails': ['test1@hello.com', 'test2@hello.com']}


class TestParamSignal(TextSignal):
  name: ClassVar[str] = 'param_signal'
  param: str

  def fields(self) -> Field:
    return field('string')

  def compute(self, data: Iterable[RichData]) -> Iterator[Optional[Item]]:
    for text_content in data:
      yield f'{str(text_content)}_{self.param}'


class TestSignal(TextSignal):
  name: ClassVar[str] = 'test_signal'

  @override
  def fields(self) -> Field:
    return field(fields={'len': 'int32', 'flen': 'float32'})

  @override
  def compute(self, data: Iterable[RichData]) -> Iterator[Optional[Item]]:
    return ({'len': len(text_content), 'flen': float(len(text_content))} for text_content in data)


class TestSplitSignal(TextSignal):
  """Split documents into sentence by splitting on period, generating entities."""

  name: ClassVar[str] = 'test_split'

  @override
  def fields(self) -> Field:
    return field(fields=['string_span'])

  @override
  def compute(self, data: Iterable[RichData]) -> Iterator[Item]:
    for text in data:
      if not isinstance(text, str):
        raise ValueError(f'Expected text to be a string, got {type(text)} instead.')
      sentences = [f'{sentence.strip()}.' for sentence in text.split('.') if sentence]
      yield [
        span(text.index(sentence), text.index(sentence) + len(sentence)) for sentence in sentences
      ]


EMBEDDINGS: list[tuple[str, Union[list[float], list[list[float]]]]] = [
  ('hello.', [1.0, 0.0, 0.0]),
  # This embedding has an outer dimension of 1.
  ('hello2.', [[1.0, 1.0, 0.0]]),
  ('hello3.', [[0, 0, 1.0]]),
]

STR_EMBEDDINGS: dict[str, Union[list[float], list[list[float]]]] = {
  text: embedding for text, embedding in EMBEDDINGS
}


class TestEmbedding(TextEmbeddingSignal):
  """A test embed function."""

  name: ClassVar[str] = 'test_embedding'

  @override
  def compute(self, data: Iterable[RichData]) -> Iterator[Item]:
    """Call the embedding function."""
    for example in data:
      example = cast(str, example)
      yield [lilac_embedding(0, len(example), np.array(STR_EMBEDDINGS[example]))]


class ComputedKeySignal(TextSignal):
  name: ClassVar[str] = 'computed_key'

  @override
  def fields(self) -> Field:
    return field('int64')

  @override
  def compute(self, data: Iterable[RichData]) -> Iterator[Optional[Item]]:
    for text in data:
      yield 1

  def key(self, is_computed_signal: Optional[bool] = False) -> str:
    return f'key_{is_computed_signal}'


@pytest.fixture(scope='module', autouse=True)
def setup_teardown() -> Iterable[None]:
  # Setup.
  clear_signal_registry()
  register_source(TestSource)
  register_signal(TestSparseSignal)
  register_signal(TestSparseRichSignal)
  register_signal(TestParamSignal)
  register_signal(TestSignal)
  register_signal(TestSplitSignal)
  register_signal(TestEmbedding)
  register_signal(ComputedKeySignal)
  register_signal(ConceptSignal)

  # Unit test runs.
  yield
  # Teardown.
  clear_source_registry()
  clear_signal_registry()


def test_signal_output_validation(make_test_data: TestDataMaker) -> None:
  class _TestInvalidSignal(TextSignal):
    name: ClassVar[str] = 'test_invalid_signal'

    @override
    def fields(self) -> Field:
      return field('int32')

    @override
    def compute(self, data: Iterable[RichData]) -> Iterator[Optional[Item]]:
      # Return an invalid output that doesn't match the input length.
      yield from ()

  signal = _TestInvalidSignal()

  dataset = make_test_data([{'text': 'hello'}, {'text': 'hello world'}])

  with pytest.raises(
    ValueError,
    match=re.escape(
      'The signal generated a different number of outputs than was given as input. '
      'Please yield `None` for sparse signals. For signals that output multiple values, '
      'please yield an array for each input.'
    ),
  ):
    dataset.compute_signal(signal, 'text')


def test_signal_error_propagates(make_test_data: TestDataMaker) -> None:
  class _TestInvalidSignal(TextSignal):
    name: ClassVar[str] = 'test_invalid_signal'

    @override
    def fields(self) -> Field:
      return field('int32')

    @override
    def compute(self, data: Iterable[RichData]) -> Iterator[Optional[Item]]:
      raise ValueError('signal error')

  signal = _TestInvalidSignal()

  dataset = make_test_data([{'text': 'hello'}, {'text': 'hello world'}])

  with pytest.raises(
    ValueError,
    match=re.escape('signal error'),
  ):
    dataset.compute_signal(signal, 'text')


def test_sparse_signal(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'hello'}, {'text': 'hello world'}])

  dataset.compute_signal(TestSparseSignal(), 'text')

  result = dataset.select_rows(['text'], combine_columns=True)
  assert list(result) == [
    {'text': enriched_item('hello', {'test_sparse_signal': None})},
    {'text': enriched_item('hello world', {'test_sparse_signal': 11})},
  ]


def test_sparse_rich_signal(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'hello'}, {'text': 'hello world'}])

  dataset.compute_signal(TestSparseRichSignal(), 'text')

  result = dataset.select_rows(['text'], combine_columns=True)
  assert list(result) == [
    {'text': enriched_item('hello', {'test_sparse_rich_signal': None})},
    {
      'text': enriched_item(
        'hello world',
        {'test_sparse_rich_signal': {'emails': ['test1@hello.com', 'test2@hello.com']}},
      )
    },
  ]


def test_signal_continuation(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(
    [
      {'text': 'a'},
      {'text': 'ab'},
      {'text': 'abc'},
    ]
  )
  orig_manifest = dataset.manifest()
  orig_rows = list(dataset.select_rows(['*']))

  first_run = True
  processed_text: list[RichData] = []

  class _TestSignal(TextSignal):
    name: ClassVar[str] = 'test_signal'

    @override
    def fields(self) -> Field:
      return field(dtype='int32')

    @override
    def compute(self, data: Iterable[RichData]) -> Iterator[Optional[Item]]:
      for i, item in enumerate(data):
        if first_run and item == 'ab':
          raise ValueError('Throwing')
        processed_text.append(item)
        yield len(item)

  test_signal = _TestSignal()
  with pytest.raises(Exception):
    dataset.compute_signal(test_signal, 'text')

  assert processed_text == ['a']

  # When the signal throws an error, nothing should change.
  assert dataset.manifest() == orig_manifest
  assert list(dataset.select_rows([PATH_WILDCARD])) == orig_rows

  first_run = False
  processed_text = []
  dataset.compute_signal(test_signal, 'text')

  assert processed_text == ['ab', 'abc']

  assert dataset.manifest() == DatasetManifest(
    namespace=TEST_NAMESPACE,
    dataset_name=TEST_DATASET_NAME,
    data_schema=schema(
      {
        'text': field(
          'string',
          fields={
            'test_signal': field(dtype='int32', signal=test_signal.model_dump(exclude_none=True))
          },
        ),
      }
    ),
    num_items=3,
    source=TestSource(),
  )
  rows = list(dataset.select_rows([PATH_WILDCARD], combine_columns=True))
  assert rows == [
    {'text': enriched_item('a', {'test_signal': 1})},
    {'text': enriched_item('ab', {'test_signal': 2})},
    {'text': enriched_item('abc', {'test_signal': 3})},
  ]


def test_signal_overwrite(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(SIMPLE_ITEMS)

  test_signal = TestSignal()

  expected_manifest = DatasetManifest(
    namespace=TEST_NAMESPACE,
    dataset_name=TEST_DATASET_NAME,
    data_schema=schema(
      {
        'str': field(
          'string',
          fields={
            'test_signal': field(
              signal=test_signal.model_dump(exclude_none=True),
              fields={'len': 'int32', 'flen': 'float32'},
            )
          },
        ),
        'int': 'int32',
        'bool': 'boolean',
        'float': 'float32',
      }
    ),
    num_items=3,
    source=TestSource(),
  )
  expected_items = [
    {'str': enriched_item('a', {'test_signal': {'len': 1, 'flen': 1.0}})},
    {'str': enriched_item('b', {'test_signal': {'len': 1, 'flen': 1.0}})},
    {'str': enriched_item('b', {'test_signal': {'len': 1, 'flen': 1.0}})},
  ]

  dataset.compute_signal(test_signal, 'str')

  # Check the enriched dataset manifest has 'text' enriched.
  assert dataset.manifest() == expected_manifest
  assert list(dataset.select_rows(['str'], combine_columns=True)) == expected_items

  with pytest.raises(ValueError, match='Signal already exists. Use overwrite=True to overwrite.'):
    dataset.compute_signal(test_signal, 'str')
  dataset.compute_signal(test_signal, 'str', overwrite=True)

  assert list(dataset.select_rows(['str'], combine_columns=True)) == expected_items


def test_signal_ignores_deleted(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(SIMPLE_ITEMS)
  assert dataset.manifest() == DatasetManifest(
    namespace=TEST_NAMESPACE,
    dataset_name=TEST_DATASET_NAME,
    data_schema=schema({'str': 'string', 'int': 'int32', 'bool': 'boolean', 'float': 'float32'}),
    num_items=3,
    source=TestSource(),
  )

  dataset.delete_rows(['1'])
  test_signal = TestSignal()
  dataset.compute_signal(test_signal, 'str')

  # Check the enriched dataset manifest has 'text' enriched.
  assert dataset.manifest() == DatasetManifest(
    namespace=TEST_NAMESPACE,
    dataset_name=TEST_DATASET_NAME,
    data_schema=schema(
      {
        'str': field(
          'string',
          fields={
            'test_signal': field(
              signal=test_signal.model_dump(exclude_none=True),
              fields={'len': 'int32', 'flen': 'float32'},
            )
          },
        ),
        'int': 'int32',
        'bool': 'boolean',
        'float': 'float32',
        '__deleted__': field(
          label='__deleted__',
          fields={
            'label': 'string',
            'created': 'timestamp',
          },
        ),
      },
    ),
    num_items=3,
    source=TestSource(),
  )

  result = dataset.select_rows(['str'], combine_columns=True, include_deleted=True)
  assert list(result) == [
    {'str': enriched_item('b', {'test_signal': {'len': 1, 'flen': 1.0}})},
    {'str': enriched_item('b', {'test_signal': {'len': 1, 'flen': 1.0}})},
    {'str': 'a'},
  ]


def test_source_joined_with_signal(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(SIMPLE_ITEMS)
  assert dataset.manifest() == DatasetManifest(
    namespace=TEST_NAMESPACE,
    dataset_name=TEST_DATASET_NAME,
    data_schema=schema({'str': 'string', 'int': 'int32', 'bool': 'boolean', 'float': 'float32'}),
    num_items=3,
    source=TestSource(),
  )

  test_signal = TestSignal()
  dataset.compute_signal(test_signal, 'str')

  # Check the enriched dataset manifest has 'text' enriched.
  assert dataset.manifest() == DatasetManifest(
    namespace=TEST_NAMESPACE,
    dataset_name=TEST_DATASET_NAME,
    data_schema=schema(
      {
        'str': field(
          'string',
          fields={
            'test_signal': field(
              signal=test_signal.model_dump(exclude_none=True),
              fields={'len': 'int32', 'flen': 'float32'},
            )
          },
        ),
        'int': 'int32',
        'bool': 'boolean',
        'float': 'float32',
      }
    ),
    num_items=3,
    source=TestSource(),
  )

  result = dataset.select_rows(['str'], combine_columns=True)
  assert list(result) == [
    {'str': enriched_item('a', {'test_signal': {'len': 1, 'flen': 1.0}})},
    {'str': enriched_item('b', {'test_signal': {'len': 1, 'flen': 1.0}})},
    {'str': enriched_item('b', {'test_signal': {'len': 1, 'flen': 1.0}})},
  ]

  # Select a specific signal leaf test_signal.flen with 'str'.
  result = dataset.select_rows(['str', ('str', 'test_signal', 'flen')])

  assert list(result) == [
    {'str': 'a', 'str.test_signal.flen': 1.0},
    {'str': 'b', 'str.test_signal.flen': 1.0},
    {'str': 'b', 'str.test_signal.flen': 1.0},
  ]

  # Select multiple signal leafs with aliasing.
  result = dataset.select_rows(
    [
      'str',
      Column(('str', 'test_signal', 'flen'), alias='flen'),
      Column(('str', 'test_signal', 'len'), alias='len'),
    ]
  )

  assert list(result) == [
    {'str': 'a', 'flen': 1.0, 'len': 1},
    {'str': 'b', 'flen': 1.0, 'len': 1},
    {'str': 'b', 'flen': 1.0, 'len': 1},
  ]


def test_parameterized_signal(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'hello'}, {'text': 'everybody'}])
  test_signal_a = TestParamSignal(param='a')
  test_signal_b = TestParamSignal(param='b')
  dataset.compute_signal(test_signal_a, 'text')
  dataset.compute_signal(test_signal_b, 'text')

  assert dataset.manifest() == DatasetManifest(
    namespace=TEST_NAMESPACE,
    dataset_name=TEST_DATASET_NAME,
    data_schema=schema(
      {
        'text': field(
          'string',
          fields={
            'param_signal(param=a)': field('string', test_signal_a.model_dump(exclude_none=True)),
            'param_signal(param=b)': field('string', test_signal_b.model_dump(exclude_none=True)),
          },
        )
      }
    ),
    num_items=2,
    source=TestSource(),
  )

  result = dataset.select_rows(['text'], combine_columns=True)
  assert list(result) == [
    {
      'text': enriched_item(
        'hello', {'param_signal(param=a)': 'hello_a', 'param_signal(param=b)': 'hello_b'}
      )
    },
    {
      'text': enriched_item(
        'everybody',
        {'param_signal(param=a)': 'everybody_a', 'param_signal(param=b)': 'everybody_b'},
      )
    },
  ]


def test_split_signal(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(
    [
      {'text': '[1, 1] first sentence. [1, 1] second sentence.'},
      {'text': 'b2 [2, 1] first sentence. [2, 1] second sentence.'},
    ]
  )

  signal = TestSplitSignal()
  dataset.compute_signal(signal, 'text')

  assert dataset.manifest() == DatasetManifest(
    namespace=TEST_NAMESPACE,
    dataset_name=TEST_DATASET_NAME,
    data_schema=schema(
      {
        'text': field(
          'string',
          fields={
            'test_split': field(
              signal=signal.model_dump(exclude_none=True), fields=[field('string_span')]
            )
          },
        )
      }
    ),
    num_items=2,
    source=TestSource(),
  )

  result = dataset.select_rows(['text'], combine_columns=True)
  expected_result = [
    {
      'text': enriched_item(
        '[1, 1] first sentence. [1, 1] second sentence.',
        {'test_split': [span(0, 22), span(23, 46)]},
      )
    },
    {
      'text': enriched_item(
        'b2 [2, 1] first sentence. [2, 1] second sentence.',
        {'test_split': [span(0, 25), span(26, 49)]},
      )
    },
  ]
  assert list(result) == expected_result


def test_signal_on_repeated_field(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': ['hello', 'everybody']}, {'text': ['hello2', 'everybody2']}])
  test_signal = TestSignal()
  # Run the signal on the repeated field.
  dataset.compute_signal(test_signal, ('text', '*'))

  # Check the enriched dataset manifest has 'text' enriched.
  assert dataset.manifest() == DatasetManifest(
    namespace=TEST_NAMESPACE,
    dataset_name=TEST_DATASET_NAME,
    data_schema=schema(
      {
        'text': field(
          fields=[
            field(
              'string',
              fields={
                'test_signal': field(
                  signal=test_signal.model_dump(exclude_none=True),
                  fields={'len': 'int32', 'flen': 'float32'},
                )
              },
            )
          ]
        )
      }
    ),
    num_items=2,
    source=TestSource(),
  )

  result = dataset.select_rows([('text', '*')], combine_columns=True)

  assert list(result) == [
    {
      'text': [
        enriched_item('hello', {'test_signal': {'len': 5, 'flen': 5.0}}),
        enriched_item('everybody', {'test_signal': {'len': 9, 'flen': 9.0}}),
      ]
    },
    {
      'text': [
        enriched_item('hello2', {'test_signal': {'len': 6, 'flen': 6.0}}),
        enriched_item('everybody2', {'test_signal': {'len': 10, 'flen': 10.0}}),
      ]
    },
  ]


def test_text_splitter(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(
    [
      {'text': '[1, 1] first sentence. [1, 1] second sentence.'},
      {'text': 'b2 [2, 1] first sentence. [2, 1] second sentence.'},
    ]
  )

  dataset.compute_signal(TestSplitSignal(), 'text')

  result = dataset.select_rows(['text'], combine_columns=True)
  expected_result = [
    {
      'text': enriched_item(
        '[1, 1] first sentence. [1, 1] second sentence.',
        {'test_split': [span(0, 22), span(23, 46)]},
      )
    },
    {
      'text': enriched_item(
        'b2 [2, 1] first sentence. [2, 1] second sentence.',
        {'test_split': [span(0, 25), span(26, 49)]},
      )
    },
  ]
  assert list(result) == expected_result


def test_embedding_signal(make_test_data: TestDataMaker, mocker: MockerFixture) -> None:
  dataset = make_test_data([{'text': 'hello.'}, {'text': 'hello2.'}])

  # Reduce the chunk size to 1 so we test iteratively writing to the embedding index.
  mocker.patch(f'{dataset_utils_module.__name__}.EMBEDDINGS_WRITE_CHUNK_SIZE', 1)

  embedding_signal = TestEmbedding()
  dataset.compute_signal(embedding_signal, 'text')

  assert dataset.manifest() == DatasetManifest(
    namespace=TEST_NAMESPACE,
    dataset_name=TEST_DATASET_NAME,
    data_schema=schema(
      {
        'text': field(
          'string',
          fields={
            'test_embedding': field(
              signal=embedding_signal.model_dump(exclude_none=True),
              fields=[field('string_span', fields={EMBEDDING_KEY: 'embedding'})],
            )
          },
        )
      }
    ),
    num_items=2,
    source=TestSource(),
  )

  result = dataset.select_rows(combine_columns=True)
  expected_result = [{'text': 'hello.'}, {'text': 'hello2.'}]
  assert list(result) == expected_result


def test_compute_embedding_over_non_string(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'hello. hello2.'}, {'text': 'hello world. hello world2.'}])

  test_splitter = TestSplitSignal()
  dataset.compute_signal(test_splitter, 'text')

  test_embedding = TestEmbedding()
  with pytest.raises(ValueError, match='Cannot compute embedding over a non-string field.'):
    dataset.compute_signal(test_embedding, ('text', 'test_split', '*'))


def test_compute_signal_over_non_string(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'hello. hello2.'}, {'text': 'hello world. hello world2.'}])

  test_splitter = TestSplitSignal()
  dataset.compute_signal(test_splitter, 'text')

  test_embedding = TestEmbedding()
  with pytest.raises(ValueError, match='Cannot compute embedding over a non-string field.'):
    dataset.compute_signal(test_embedding, ('text', 'test_split', '*'))


def test_is_computed_signal_key(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'hello.'}, {'text': 'hello2.'}])

  signal = ComputedKeySignal()
  dataset.compute_signal(signal, 'text')

  assert dataset.manifest() == DatasetManifest(
    namespace=TEST_NAMESPACE,
    dataset_name=TEST_DATASET_NAME,
    data_schema=schema(
      {
        'text': field(
          'string', fields={'key_True': field('int64', signal=signal.model_dump(exclude_none=True))}
        )
      }
    ),
    num_items=2,
    source=TestSource(),
  )

  result = dataset.select_rows(combine_columns=True)

  expected_result = [
    {'text': enriched_item('hello.', {'key_True': 1})},
    {'text': enriched_item('hello2.', {'key_True': 1})},
  ]
  assert list(result) == expected_result


def test_concept_signal_with_select_groups(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'hello.'}, {'text': 'hello2.'}, {'text': 'hello3.'}])

  embedding_signal = TestEmbedding()
  dataset.compute_signal(embedding_signal, 'text')

  concept_db = DiskConceptDB()
  concept_db.create(namespace='test_namespace', name='test_concept', type=SignalInputType.TEXT)
  concept_db.edit(
    'test_namespace',
    'test_concept',
    ConceptUpdate(
      insert=[
        ExampleIn(label=False, text='hello.'),
        ExampleIn(label=True, text='hello2.'),
        ExampleIn(label=False, text='hello3.'),
      ]
    ),
  )

  dataset.compute_concept(
    namespace='test_namespace', concept_name='test_concept', embedding='test_embedding', path='text'
  )

  concept_key = 'test_namespace/test_concept/test_embedding'
  result = dataset.select_groups(f'text.{concept_key}.*.score')
  assert result.counts == [('Not in concept', 2), ('In concept', 1)]

  result = dataset.select_groups(
    f'text.{concept_key}.*.score', sort_by=GroupsSortBy.COUNT, sort_order=SortOrder.ASC
  )
  assert result.counts == [('In concept', 1), ('Not in concept', 2)]


def test_optional_schema(make_test_data: TestDataMaker) -> None:
  class TestSignal(TextSignal):
    name: ClassVar[str] = 'len_of_text'

    @override
    def compute(self, data: Iterable[RichData]) -> Iterator[Optional[Item]]:
      for text in data:
        yield len(text)

  dataset = make_test_data([{'text': 'hello'}, {'text': 'hello world'}])
  dataset.compute_signal(TestSignal(), 'text')
  result = dataset.select_rows(['text'], combine_columns=True)
  assert list(result) == [
    {'text': enriched_item('hello', {'len_of_text': 5})},
    {'text': enriched_item('hello world', {'len_of_text': 11})},
  ]


def test_optional_schema_parameterized_signal(make_test_data: TestDataMaker) -> None:
  class TestSignal(TextSignal):
    name: ClassVar[str] = 'len_of_text'

    param: str

    @override
    def compute(self, data: Iterable[RichData]) -> Iterator[Optional[Item]]:
      for text in data:
        yield {'len': len(text), 'first_and_last': [text[0], text[-1]]}

  dataset = make_test_data([{'text': 'hello'}, {'text': 'hello world'}])
  dataset.compute_signal(TestSignal(param='a'), 'text')
  result = dataset.select_rows(['text'], combine_columns=True)
  assert list(result) == [
    {
      'text': enriched_item(
        'hello',
        {
          'len_of_text(param=a)': {
            'len': 5,
            'first_and_last': ['h', 'o'],
          }
        },
      )
    },
    {
      'text': enriched_item(
        'hello world',
        {
          'len_of_text(param=a)': {
            'len': 11,
            'first_and_last': ['h', 'd'],
          }
        },
      )
    },
  ]



"""An interface for OpenAI chat completion generators."""


from typing import Optional

import instructor
import tiktoken
from instructor import OpenAISchema
from pydantic import Field
from typing_extensions import override

from ..env import env
from ..text_gen import TextGenerator


class OpenAIChatCompletionGenerator(TextGenerator):
  """An interface for OpenAI chat completion."""

  model: str = 'gpt-3.5-turbo-0613'
  response_description: str = ''

  @override
  def generate(self, prompt: str) -> str:
    """Generate a completion for a prompt."""
    api_key = env('OPENAI_API_KEY')
    api_type = env('OPENAI_API_TYPE')
    api_version = env('OPENAI_API_VERSION')
    if not api_key:
      raise ValueError('`OPENAI_API_KEY` environment variable not set.')

    try:
      import openai
    except ImportError:
      raise ImportError(
        'Could not import the "openai" python package. '
        'Please install it with `pip install openai`.'
      )
    openai.api_key = api_key
    if api_type:
      openai.api_type = api_type
      openai.api_version = api_version

    # Enables response_model in the openai client.
    client = instructor.patch(openai.OpenAI())

    class Completion(OpenAISchema):
      """Generated completion of a prompt."""

      completion: str = Field(..., description=self.response_description)

    return client.chat.completions.create(
      model='gpt-3.5-turbo',
      response_model=Completion,
      messages=[
        {
          'role': 'system',
          'content': 'You must call the `Completion` function with the generated completion.',
        },
        {'role': 'user', 'content': prompt},
      ],
    ).completion

  @override
  def num_tokens(self, prompt: str) -> Optional[int]:
    """Count the number of tokens for a prompt."""
    return len(tiktoken.encoding_for_model(self.model).encode(prompt))



"""Manage FastAPI background tasks."""

import functools
import uuid
from dataclasses import dataclass
from datetime import datetime, timedelta
from enum import Enum
from threading import Thread
from typing import (
  Any,
  Awaitable,
  Callable,
  Iterator,
  Literal,
  Optional,
  TypeVar,
  Union,
)

from pydantic import BaseModel
from tqdm import tqdm

from .env import env
from .utils import pretty_timedelta

TaskId = str
TaskFn = Union[Callable[..., Any], Callable[..., Awaitable[Any]]]


class TaskStatus(str, Enum):
  """Enum holding a tasks status."""

  PENDING = 'pending'
  COMPLETED = 'completed'
  ERROR = 'error'


class TaskType(str, Enum):
  """Enum holding a task type."""

  DATASET_LOAD = 'dataset_load'
  DATASET_MAP = 'dataset_map'


TaskExecutionType = Literal['processes', 'threads']


@dataclass
class TaskInfo:
  """Metadata about a task."""

  name: str
  start_timestamp: str
  status: TaskStatus
  end_timestamp: Optional[str] = None
  type: Optional[TaskType] = None
  message: Optional[str] = None
  details: Optional[str] = None

  description: Optional[str] = None
  error: Optional[str] = None

  total_len: Optional[int] = None
  total_progress: Optional[int] = None


class TaskManifest(BaseModel):
  """Information for tasks that are running or completed."""

  tasks: dict[str, TaskInfo]
  progress: Optional[float] = None


class TaskManager:
  """Manage FastAPI background tasks."""

  _tasks: dict[TaskId, TaskInfo]

  def __init__(self) -> None:
    # Maps a task id to the current progress of that task. Shared across all processes.
    self._tasks = {}

  def get_task_info(self, task_id: TaskId) -> TaskInfo:
    """Get the task info for a task."""
    return self._tasks[task_id]

  def manifest(self) -> TaskManifest:
    """Get all tasks."""
    tasks_with_progress = [
      (task.total_progress / task.total_len)
      for task in self._tasks.values()
      if task.total_progress and task.total_len and task.status != TaskStatus.COMPLETED
    ]
    return TaskManifest(
      tasks=self._tasks,
      progress=sum(tasks_with_progress) / len(tasks_with_progress) if tasks_with_progress else None,
    )

  def task_id(
    self,
    name: str,
    type: Optional[TaskType] = None,
    description: Optional[str] = None,
    total_len: Optional[int] = None,
  ) -> TaskId:
    """Create a unique ID for a task."""
    task_id = uuid.uuid4().hex
    new_task = TaskInfo(
      name=name,
      type=type,
      status=TaskStatus.PENDING,
      description=description,
      start_timestamp=datetime.now().isoformat(),
      total_len=total_len,
    )
    self._tasks[task_id] = new_task
    return task_id

  def report_progress(self, task_id: TaskId, progress: int) -> None:
    """Report the progress of a task."""
    task = self._tasks[task_id]
    task.total_progress = progress
    elapsed_sec = (datetime.now() - datetime.fromisoformat(task.start_timestamp)).total_seconds()
    ex_per_sec = progress / elapsed_sec if elapsed_sec else 0
    # 1748/1748 [elapsed 00:16<00:00, 106.30 ex/s]
    elapsed = pretty_timedelta(timedelta(seconds=elapsed_sec))
    task.details = f'{progress:,}/{task.total_len:,} [{elapsed} {ex_per_sec:,.2f} ex/s]'

  def set_error(self, task_id: TaskId, error: str) -> None:
    """Mark a task as errored."""
    task = self._tasks[task_id]
    task.status = TaskStatus.ERROR
    task.error = error
    task.end_timestamp = datetime.now().isoformat()

  def set_completed(self, task_id: TaskId) -> None:
    """Mark a task completed."""
    end_timestamp = datetime.now().isoformat()
    task = self._tasks[task_id]
    task.end_timestamp = end_timestamp

    elapsed = datetime.fromisoformat(end_timestamp) - datetime.fromisoformat(task.start_timestamp)
    elapsed_formatted = pretty_timedelta(elapsed)

    if task.status != TaskStatus.ERROR:
      task.status = TaskStatus.COMPLETED
      task.message = f'Completed in {elapsed_formatted}'


@functools.cache
def get_task_manager() -> TaskManager:
  """The global singleton for the task manager."""
  return TaskManager()


TProgress = TypeVar('TProgress')


def get_progress_bar(
  task_id: Optional[TaskId] = None,
  task_description: Optional[str] = None,
  estimated_len: Optional[int] = None,
  offset: int = 0,
) -> Callable[[Iterator[TProgress]], Iterator[TProgress]]:
  """An iterable wrapper that emits progress and yields the original iterable."""
  # reduce unit test spam
  if env('LILAC_TEST', False):
    return lambda it: it
  if task_id is None:
    return lambda it: iter(tqdm(it, initial=offset, total=estimated_len, desc=task_description))

  task_manager = get_task_manager()
  task_info = task_manager.get_task_info(task_id)
  if estimated_len:
    task_info.total_len = estimated_len
  if task_description:
    task_info.description = task_description

  def progress_reporter(it: Iterator[TProgress]) -> Iterator[TProgress]:
    progress = offset
    try:
      for item in tqdm(it, initial=progress, total=task_info.total_len, desc=task_info.description):
        progress += 1
        if progress % 100 == 0:
          task_manager.report_progress(task_id, progress)
        yield item
    except Exception as e:
      task_manager.set_error(task_id, str(e))
      raise e
    task_manager.set_completed(task_id)

  return progress_reporter


def launch_task(task_id: TaskId, run_fn: Callable) -> None:
  """Launch a task in a thread, handling exit conditions, etc.."""
  tm = get_task_manager()

  def _wrapper() -> None:
    try:
      run_fn()
    except Exception as e:
      tm.set_error(task_id, str(e))
    else:
      tm.set_completed(task_id)

  thread = Thread(target=_wrapper, daemon=True)
  thread.start()



"""Tests for load_dataset.py."""

import os
import pathlib
import uuid
from typing import ClassVar, Iterable, Optional, Type

import pyarrow as pa
import pyarrow.parquet as pq
import pytest
from datasets import Dataset, DatasetDict
from pytest_mock import MockerFixture
from typing_extensions import override

from .config import Config, DatasetConfig, DatasetSettings, DatasetUISettings
from .data.dataset_duckdb import read_source_manifest
from .data.dataset_utils import get_parquet_filename, schema_to_arrow_schema
from .load_dataset import from_dicts, from_huggingface, process_source
from .project import read_project_config
from .schema import PARQUET_FILENAME_PREFIX, ROWID, STRING, Field, Item, Schema, schema
from .source import Source, SourceManifest, SourceSchema, clear_source_registry, register_source
from .tasks import TaskId
from .test_utils import fake_uuid, retrieve_parquet_rows
from .utils import DATASETS_DIR_NAME


class TestSource(Source):
  """A test source."""

  name: ClassVar[str] = 'test_source'

  @override
  def setup(self) -> None:
    pass

  @override
  def source_schema(self) -> SourceSchema:
    """Return the source schema."""
    return SourceSchema(fields=schema({'x': 'int64', 'y': 'string'}).fields, num_items=2)

  @override
  def yield_items(self) -> Iterable[Item]:
    return [{'x': 1, 'y': 'ten'}, {'x': 2, 'y': 'twenty'}]


class TestFastSource(Source):
  """A fast test source."""

  name: ClassVar[str] = 'test_fast_source'

  @override
  def setup(self) -> None:
    pass

  @override
  def source_schema(self) -> SourceSchema:
    """Return the source schema."""
    return SourceSchema(fields=schema({'x': 'int64', 'y': 'string'}).fields, num_items=2)

  @override
  def load_to_parquet(self, output_dir: str, task_id: Optional[TaskId] = None) -> SourceManifest:
    rows = [
      {ROWID: fake_uuid(b'1').hex, 'x': 1, 'y': 'ten'},
      {ROWID: fake_uuid(b'2').hex, 'x': 2, 'y': 'twenty'},
    ]
    schema = Schema(fields=self.source_schema().fields.copy())
    schema.fields[ROWID] = Field(dtype=STRING)

    arrow_schema = schema_to_arrow_schema(schema)
    table = pa.Table.from_pylist(rows, schema=arrow_schema)

    filename = get_parquet_filename(PARQUET_FILENAME_PREFIX, 0, 1)
    out_file = os.path.join(output_dir, filename)
    os.makedirs(output_dir, exist_ok=True)
    pq.write_table(table, out_file)

    return SourceManifest(
      files=[filename], data_schema=Schema(fields=self.source_schema().fields), source=self
    )


@pytest.fixture(scope='module', autouse=True)
def setup_teardown() -> Iterable[None]:
  # Setup.
  register_source(TestSource)
  register_source(TestFastSource)
  # Unit test runs.
  yield
  # Teardown.
  clear_source_registry()


@pytest.fixture(autouse=True)
def set_project_dir(tmp_path: pathlib.Path, mocker: MockerFixture) -> None:
  mocker.patch.dict(os.environ, {'LILAC_PROJECT_DIR': str(tmp_path)})


@pytest.mark.parametrize('source_cls', [TestSource, TestFastSource])
def test_data_loader(
  source_cls: Type[Source], tmp_path: pathlib.Path, mocker: MockerFixture
) -> None:
  mock_uuid = mocker.patch.object(uuid, 'uuid4', autospec=True)
  mock_uuid.side_effect = [fake_uuid(b'1'), fake_uuid(b'2')]

  source = source_cls()
  setup_mock = mocker.spy(source_cls, 'setup')

  output_dir = process_source(
    tmp_path,
    DatasetConfig(namespace='test_namespace', name='test_dataset', source=source),
  )

  assert setup_mock.call_count == 1

  assert output_dir == os.path.join(tmp_path, DATASETS_DIR_NAME, 'test_namespace', 'test_dataset')

  source_manifest = read_source_manifest(output_dir)

  assert source_manifest == SourceManifest(
    files=[get_parquet_filename(PARQUET_FILENAME_PREFIX, 0, 1)],
    data_schema=schema({'x': 'int64', 'y': 'string'}),
    source=source,
  )

  items = retrieve_parquet_rows(pathlib.Path(output_dir), source_manifest, retain_rowid=True)

  assert items == [
    {ROWID: fake_uuid(b'1').hex, 'x': 1, 'y': 'ten'},
    {ROWID: fake_uuid(b'2').hex, 'x': 2, 'y': 'twenty'},
  ]

  project_config = read_project_config(str(tmp_path))
  assert project_config == Config(
    datasets=[
      DatasetConfig(
        namespace='test_namespace',
        name='test_dataset',
        source=source,
        # 'y' is the longest path, so should be set as the default setting.
        settings=DatasetSettings(ui=DatasetUISettings(media_paths=[('y',)])),
      )
    ]
  )


def test_from_dicts() -> None:
  items = [{'text': 'hello'}, {'text': 'world'}]
  ds = from_dicts('local', 'test', items)
  assert list(ds.select_rows()) == items


def test_from_hf_dataset() -> None:
  def gen() -> Iterable[dict[str, str]]:
    yield {'text': 'hello'}
    yield {'text': 'world'}

  hf_ds = Dataset.from_generator(gen)
  ds = from_huggingface(hf_ds)
  assert list(ds.select_rows()) == [
    {'__hfsplit__': 'default', 'text': 'hello'},
    {'__hfsplit__': 'default', 'text': 'world'},
  ]
  assert ds.namespace == 'local'
  assert ds.dataset_name == 'generator'


def test_from_hf_dataset_dict() -> None:
  def gen() -> Iterable[dict[str, str]]:
    yield {'text': 'hello'}
    yield {'text': 'world'}

  hf_ds_dict = DatasetDict({'train': Dataset.from_generator(gen)})
  ds = from_huggingface(hf_ds_dict)
  assert list(ds.select_rows()) == [
    {'__hfsplit__': 'train', 'text': 'hello'},
    {'__hfsplit__': 'train', 'text': 'world'},
  ]
  assert ds.namespace == 'local'
  assert ds.dataset_name == 'generator'


def test_from_hf_explicit_namespace_and_name() -> None:
  def gen() -> Iterable[dict[str, str]]:
    yield {'text': 'hello'}
    yield {'text': 'world'}

  hf_ds = Dataset.from_generator(gen)
  ds = from_huggingface(hf_ds, namespace='local', name='test')
  assert list(ds.select_rows()) == [
    {'__hfsplit__': 'default', 'text': 'hello'},
    {'__hfsplit__': 'default', 'text': 'world'},
  ]
  assert ds.namespace == 'local'
  assert ds.dataset_name == 'test'



"""Registers all available default signals."""

from ..embeddings.cohere import Cohere
from ..embeddings.gte import GTEBase, GTESmall, GTETiny
from ..embeddings.jina import JinaV2Base, JinaV2Small
from ..embeddings.openai import OpenAIEmbedding
from ..embeddings.palm import PaLM
from ..embeddings.sbert import SBERT
from ..signal import register_signal
from .concept_labels import ConceptLabelsSignal
from .concept_scorer import ConceptSignal
from .lang_detection import LangDetectionSignal
from .markdown_code_block import MarkdownCodeBlockSignal
from .near_dup import NearDuplicateSignal
from .ner import SpacyNER
from .pii import PIISignal
from .text_statistics import TextStatisticsSignal


def register_default_signals() -> None:
  """Register all the default signals."""
  # Concepts.
  register_signal(ConceptSignal)
  register_signal(ConceptLabelsSignal)

  # Text.
  register_signal(PIISignal)
  register_signal(TextStatisticsSignal)
  register_signal(SpacyNER)
  register_signal(NearDuplicateSignal)
  register_signal(LangDetectionSignal)
  register_signal(MarkdownCodeBlockSignal)

  # Embeddings.
  register_signal(Cohere)

  register_signal(SBERT)

  register_signal(OpenAIEmbedding)

  register_signal(PaLM)

  register_signal(GTETiny)
  register_signal(GTESmall)
  register_signal(GTEBase)

  register_signal(JinaV2Small)
  register_signal(JinaV2Base)



"""A signal to search for a substring in a document."""
from typing import ClassVar, Iterable, Iterator, Optional

from typing_extensions import override

from ..schema import Field, Item, RichData, SignalInputType, field, span
from ..signal import Signal


def _find_all(text: str, subtext: str) -> Iterable[tuple[int, int]]:
  # Ignore casing.
  text = text.lower()
  subtext = subtext.lower()
  subtext_len = len(subtext)
  start = 0
  while True:
    start = text.find(subtext, start)
    if start == -1:
      return
    yield start, start + subtext_len
    start += subtext_len


class SubstringSignal(Signal):
  """Find a substring in a document."""

  name: ClassVar[str] = 'substring_search'
  display_name: ClassVar[str] = 'Substring Search'
  input_type: ClassVar[SignalInputType] = SignalInputType.TEXT

  query: str

  @override
  def fields(self) -> Field:
    return field(fields=['string_span'])

  @override
  def compute(self, data: Iterable[RichData]) -> Iterator[Optional[Item]]:
    for text in data:
      if not isinstance(text, str):
        yield None
        continue
      yield [span(start, end) for start, end in _find_all(text, self.query)]



"""Interface for storing vectors."""

import abc
import os
import pickle
from typing import Iterable, Optional, Sequence, Type, cast

import numpy as np

from ..schema import SpanVector, VectorKey
from ..utils import open_file


class VectorStore(abc.ABC):
  """Interface for storing and retrieving vectors."""

  # The global name of the vector store.
  name: str

  @abc.abstractmethod
  def save(self, base_path: str) -> None:
    """Save the store to disk."""
    pass

  @abc.abstractmethod
  def load(self, base_path: str) -> None:
    """Load the store from disk."""
    pass

  @abc.abstractmethod
  def size(self) -> int:
    """Return the number of vectors in the store."""
    pass

  @abc.abstractmethod
  def add(self, keys: list[VectorKey], embeddings: np.ndarray) -> None:
    """Add or edit the given keyed embeddings to the store.

    If the keys already exist they will be overwritten, acting as an "upsert".

    Args:
      keys: The keys to add the embeddings for.
      embeddings: The embeddings to add. This should be a 2D matrix with the same length as keys.
    """
    pass

  @abc.abstractmethod
  def get(self, keys: Optional[Iterable[VectorKey]] = None) -> np.ndarray:
    """Return the embeddings for given keys.

    Args:
      keys: The keys to return the embeddings for. If None, return all embeddings.

    Returns:
      The embeddings for the given keys.
    """
    pass

  def topk(
    self, query: np.ndarray, k: int, keys: Optional[Iterable[VectorKey]] = None
  ) -> list[tuple[VectorKey, float]]:
    """Return the top k most similar vectors.

    Args:
      query: The query vector.
      k: The number of results to return.
      keys: Optional keys to restrict the search to.

    Returns:
      A list of (key, score) tuples.
    """
    raise NotImplementedError


PathKey = VectorKey

_SPANS_PICKLE_NAME = 'spans.pkl'


class VectorDBIndex:
  """Stores and retrives span vectors.

  This wraps a regular vector store by adding a mapping from path keys, such as (rowid1, 0),
  to span keys, such as (rowid1, 0, 0), which denotes the first span in the (rowid1, 0) document.
  """

  def __init__(self, vector_store: str) -> None:
    self._vector_store: VectorStore = get_vector_store_cls(vector_store)()
    # Map a path key to spans for that path.
    self._id_to_spans: dict[PathKey, list[tuple[int, int]]] = {}
    self._rowid_to_path_keys: dict[str, list[PathKey]] = {}

  def load(self, base_path: str) -> None:
    """Load the vector index from disk."""
    assert not self._id_to_spans, 'Cannot load into a non-empty index.'
    with open_file(os.path.join(base_path, _SPANS_PICKLE_NAME), 'rb') as f:
      all_spans: list[tuple[PathKey, list[tuple[int, int]]]] = pickle.load(f)
      self._id_to_spans.update(all_spans)
      for path_key, _ in all_spans:
        rowid = cast(str, path_key[0])
        self._rowid_to_path_keys.setdefault(rowid, []).append(path_key)
    self._vector_store.load(os.path.join(base_path, self._vector_store.name))

  def save(self, base_path: str) -> None:
    """Save the vector index to disk."""
    assert self._id_to_spans, 'Cannot save an empty index.'
    with open_file(os.path.join(base_path, _SPANS_PICKLE_NAME), 'wb') as f:
      pickle.dump(list(self._id_to_spans.items()), f)
    self._vector_store.save(os.path.join(base_path, self._vector_store.name))

  def add(
    self, all_spans: Sequence[tuple[PathKey, list[tuple[int, int]]]], embeddings: np.ndarray
  ) -> None:
    """Add the given spans and embeddings.

    Args:
      all_spans: The spans to initialize the index with.
      embeddings: The embeddings to initialize the index with.
    """
    vector_keys = [(*path_key, i) for path_key, spans in all_spans for i in range(len(spans))]
    assert len(vector_keys) == len(
      embeddings
    ), f'Number of spans ({len(vector_keys)}) and embeddings ({len(embeddings)}) must match.'

    self._id_to_spans.update(all_spans)
    for path_key, _ in all_spans:
      rowid = cast(str, path_key[0])
      self._rowid_to_path_keys.setdefault(rowid, []).append(path_key)

    self._vector_store.add(vector_keys, embeddings)

  def get_vector_store(self) -> VectorStore:
    """Return the underlying vector store."""
    return self._vector_store

  def get(self, keys: Iterable[PathKey]) -> Iterable[list[SpanVector]]:
    """Return the spans with vectors for each key in `keys`.

    Args:
      keys: The keys to return the vectors for.

    Returns:
      The span vectors for the given keys.
    """
    all_spans: list[list[tuple[int, int]]] = []
    all_vector_keys: list[list[VectorKey]] = []
    for path_key in keys:
      spans = self._id_to_spans.get(path_key, [])
      all_spans.append(spans)
      all_vector_keys.append([(*path_key, i) for i in range(len(spans))])

    offset = 0
    flat_vector_keys = [key for vector_keys in all_vector_keys for key in (vector_keys or [])]
    all_vectors = self._vector_store.get(flat_vector_keys)
    for spans in all_spans:
      vectors = all_vectors[offset : offset + len(spans)]
      yield [{'span': span, 'vector': vector} for span, vector in zip(spans, vectors)]
      offset += len(spans)

  def topk(
    self, query: np.ndarray, k: int, rowids: Optional[Iterable[str]] = None
  ) -> list[tuple[PathKey, float]]:
    """Return the top k most similar vectors.

    Args:
      query: The query vector.
      k: The number of results to return.
      rowids: Optional row ids to restrict the search to.

    Returns:
      A list of (rowid, score) tuples.
    """
    total_num_span_keys = self._vector_store.size()
    k = min(k, total_num_span_keys)
    span_keys: Optional[list[VectorKey]] = None
    if rowids is not None:
      span_keys = []
      for rowid in rowids:
        path_keys = self._rowid_to_path_keys.get(rowid)
        if path_keys is None:
          continue

        span_keys.extend(
          [
            (*path_key, i)
            for path_key in path_keys
            for i in range(len(self._id_to_spans[path_key]))
          ]
        )
      k = min(k, len(span_keys))
    span_k = k
    path_key_scores: dict[PathKey, float] = {}
    seen_rowids: dict[str, bool] = {}
    while (
      len(seen_rowids) < k
      and span_k <= total_num_span_keys
      and (not span_keys or span_k <= len(span_keys))
    ):
      span_k += k
      vector_key_scores = self._vector_store.topk(query, span_k, span_keys)
      for (*path_key_list, _), score in vector_key_scores:
        path_key = tuple(path_key_list)
        if path_key not in path_key_scores:
          path_key_scores[path_key] = score
        rowid = cast(str, path_key[0])
        if rowid not in seen_rowids:
          seen_rowids[rowid] = True

    top_rowids = set(list(seen_rowids.keys())[:k])
    top_path_keys = [(key, s) for (key, s) in path_key_scores.items() if key[0] in top_rowids]
    return list(top_path_keys)


VECTOR_STORE_REGISTRY: dict[str, Type[VectorStore]] = {}


def register_vector_store(vector_store_cls: Type[VectorStore]) -> None:
  """Register a vector store in the global registry."""
  if vector_store_cls.name in VECTOR_STORE_REGISTRY:
    raise ValueError(f'Vector store "{vector_store_cls.name}" has already been registered!')

  VECTOR_STORE_REGISTRY[vector_store_cls.name] = vector_store_cls


def get_vector_store_cls(vector_store_name: str) -> Type[VectorStore]:
  """Return a registered vector store given the name in the registry."""
  return VECTOR_STORE_REGISTRY[vector_store_name]


def clear_vector_store_registry() -> None:
  """Clear the vector store registry."""
  VECTOR_STORE_REGISTRY.clear()



"""Writes the openapi.json file to the specified output.

This is meant to run as a standalone script. It lives in lilac/ so we can import the FastAPI app.
"""
import json

import click
from fastapi.openapi.utils import get_openapi

from .server import app


@click.command()
@click.option(
  '--output', required=True, type=str, help='The output filepath for the opepnapi.json file.'
)
def main(output: str) -> None:
  """Create the openapi.json file for the API to generate TypeScript stubs."""
  with open(output, 'w') as f:
    json.dump(
      get_openapi(
        title=app.title,
        version=app.version,
        openapi_version=app.openapi_version,
        separate_input_output_schemas=False,
        description=app.description,
        routes=app.routes,
      ),
      f,
    )


if __name__ == '__main__':
  main()



"""Tests for the paquet source."""

import os
import pathlib

import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
import pytest
from pydantic import ValidationError

from ..config import DatasetConfig
from ..load_dataset import create_dataset
from ..project import create_project_and_set_env
from ..schema import STRING, Field, MapType, field, schema
from ..source import SourceSchema
from ..test_utils import retrieve_parquet_rows
from ..utils import chunks
from .parquet_source import ParquetSource


def test_simple_rows(tmp_path: pathlib.Path) -> None:
  table = pa.Table.from_pylist(
    [{'name': 'a', 'age': 1}, {'name': 'b', 'age': 2}, {'name': 'c', 'age': 3}]
  )

  out_file = os.path.join(tmp_path, 'test.parquet')
  pq.write_table(table, out_file)

  source = ParquetSource(filepaths=[out_file])
  source.setup()
  source_schema = source.source_schema()
  assert source_schema == SourceSchema(
    fields=schema({'name': 'string', 'age': 'int64'}).fields, num_items=None
  )
  manifest = source.load_to_parquet(str(tmp_path), task_id=None)
  items = retrieve_parquet_rows(tmp_path, manifest)
  assert items == [{'name': 'a', 'age': 1}, {'name': 'b', 'age': 2}, {'name': 'c', 'age': 3}]


def test_map_dtype(tmp_path: pathlib.Path) -> None:
  # Create data for the table
  data = {
    'column': [
      {'a': 1.0, 'b': 2.0},
      {'b': 2.5},
      {'a': 3.0, 'c': 3.5},
    ]
  }
  # Create a table with a single column of type map<string, float>.
  table = pa.table(data, schema=pa.schema([('column', pa.map_(pa.string(), pa.float32()))]))

  out_file = os.path.join(tmp_path, 'test.parquet')
  pq.write_table(table, out_file)

  source = ParquetSource(filepaths=[out_file])
  source.setup()
  source_schema = source.source_schema()
  assert source_schema == SourceSchema(
    fields=schema({'column': field(MapType(key_type=STRING, value_field=field('float32')))}).fields,
    num_items=None,
  )
  manifest = source.load_to_parquet(str(tmp_path), task_id=None)
  items = retrieve_parquet_rows(tmp_path, manifest)
  assert items == [
    {'column': [('a', 1.0), ('b', 2.0)]},
    {'column': [('b', 2.5)]},
    {'column': [('a', 3.0), ('c', 3.5)]},
  ]


def test_single_shard_with_sampling(tmp_path: pathlib.Path) -> None:
  source_items = [{'name': 'a', 'age': 1}, {'name': 'b', 'age': 2}, {'name': 'c', 'age': 3}]
  table = pa.Table.from_pylist(source_items)

  out_file = os.path.join(tmp_path, 'test.parquet')
  pq.write_table(table, out_file)

  # Test sampling with different sample sizes, including sample size > num_items.
  for sample_size in range(1, 5):
    source = ParquetSource(filepaths=[out_file], sample_size=sample_size)
    source.setup()
    manifest = source.load_to_parquet(str(tmp_path), task_id=None)
    items = retrieve_parquet_rows(tmp_path, manifest)
    assert len(items) == min(sample_size, len(source_items))


def test_single_shard_pseudo_shuffle(tmp_path: pathlib.Path) -> None:
  source_items = [{'name': 'a', 'age': 1}, {'name': 'b', 'age': 2}, {'name': 'c', 'age': 3}]
  table = pa.Table.from_pylist(source_items)

  out_file = os.path.join(tmp_path, 'test.parquet')
  pq.write_table(table, out_file)

  # Test sampling with different sample sizes, including sample size > num_items.
  for sample_size in range(1, 5):
    source = ParquetSource(filepaths=[out_file], sample_size=sample_size, pseudo_shuffle=True)
    source.setup()
    manifest = source.load_to_parquet(str(tmp_path), task_id=None)
    items = retrieve_parquet_rows(tmp_path, manifest)
    assert len(items) == min(sample_size, len(source_items))


def test_multi_shard(tmp_path: pathlib.Path) -> None:
  source_items = [{'name': 'a', 'age': 1}, {'name': 'b', 'age': 2}, {'name': 'c', 'age': 3}]
  for i, item in enumerate(source_items):
    table = pa.Table.from_pylist([item])
    out_file = tmp_path / f'test-{i}.parquet'
    pq.write_table(table, out_file)

  source = ParquetSource(filepaths=[str(tmp_path / 'test-*.parquet')])
  source.setup()
  manifest = source.load_to_parquet(str(tmp_path), task_id=None)
  items = retrieve_parquet_rows(tmp_path, manifest)
  items.sort(key=lambda x: x['name'])
  assert items == source_items


def test_multi_shard_sample(tmp_path: pathlib.Path) -> None:
  source_items = [{'name': 'a', 'age': 1}, {'name': 'b', 'age': 2}, {'name': 'c', 'age': 3}]
  for i, item in enumerate(source_items):
    table = pa.Table.from_pylist([item])
    out_file = tmp_path / f'test-{i}.parquet'
    pq.write_table(table, out_file)

  # Test sampling with different sample sizes, including sample size > num_items.
  for sample_size in range(1, 5):
    source = ParquetSource(filepaths=[str(tmp_path / 'test-*.parquet')], sample_size=sample_size)
    source.setup()
    manifest = source.load_to_parquet(str(tmp_path), task_id=None)
    items = retrieve_parquet_rows(tmp_path, manifest)
    assert len(items) == min(sample_size, len(source_items))


def test_multi_shard_approx_shuffle(tmp_path: pathlib.Path) -> None:
  source_items = [{'name': 'a', 'age': 1}, {'name': 'b', 'age': 2}, {'name': 'c', 'age': 3}]
  for i, item in enumerate(source_items):
    table = pa.Table.from_pylist([item])
    out_file = tmp_path / f'test-{i}.parquet'
    pq.write_table(table, out_file)

  # Test sampling with different sample sizes, including sample size > num_items.
  for sample_size in range(1, 5):
    source = ParquetSource(
      filepaths=[str(tmp_path / 'test-*.parquet')],
      pseudo_shuffle=True,
      sample_size=sample_size,
    )
    source.setup()
    manifest = source.load_to_parquet(str(tmp_path), task_id=None)
    items = retrieve_parquet_rows(tmp_path, manifest)
    assert len(items) == min(sample_size, len(source_items))


def test_uniform_shards_pseudo_shuffle(tmp_path: pathlib.Path) -> None:
  source_items = [{'index': i} for i in range(100)]
  for i, chunk in enumerate(chunks(source_items, 10)):
    table = pa.Table.from_pylist(chunk)
    out_file = tmp_path / f'test-{i}.parquet'
    pq.write_table(table, out_file)

  source = ParquetSource(
    filepaths=[str(tmp_path / 'test-*.parquet')], pseudo_shuffle=True, sample_size=20
  )
  source.setup()
  manifest = source.load_to_parquet(str(tmp_path), task_id=None)
  items = retrieve_parquet_rows(tmp_path, manifest)
  assert len(items) == 20


def test_nonuniform_shards_pseudo_shuffle(tmp_path: pathlib.Path) -> None:
  source_items = [{'index': i} for i in range(100)]
  shard_sizes = [49, 1, 40, 10]
  for i, shard_size in enumerate(shard_sizes):
    chunk = source_items[:shard_size]
    source_items = source_items[shard_size:]
    table = pa.Table.from_pylist(chunk)
    out_file = tmp_path / f'test-{i}.parquet'
    pq.write_table(table, out_file)

  source = ParquetSource(
    filepaths=[str(tmp_path / 'test-*.parquet')], pseudo_shuffle=True, sample_size=20
  )
  source.setup()
  manifest = source.load_to_parquet(str(tmp_path), task_id=None)
  items = retrieve_parquet_rows(tmp_path, manifest)
  assert len(items) == 20


def test_sampling_with_seed(tmp_path: pathlib.Path) -> None:
  source_items = [{'index': i} for i in range(100)]
  for i, chunk in enumerate(chunks(source_items, 10)):
    table = pa.Table.from_pylist(chunk)
    out_file = tmp_path / f'test-{i}.parquet'
    pq.write_table(table, out_file)

  source = ParquetSource(filepaths=[str(tmp_path / 'test-*.parquet')], sample_size=20, seed=42)
  source.setup()
  manifest = source.load_to_parquet(str(tmp_path), task_id=None)
  items = retrieve_parquet_rows(tmp_path, manifest)
  assert len(items) == 20


def test_approx_shuffle_with_seed(tmp_path: pathlib.Path) -> None:
  source_items = [{'index': i} for i in range(100)]
  for i, chunk in enumerate(chunks(source_items, 10)):
    table = pa.Table.from_pylist(chunk)
    out_file = tmp_path / f'test-{i}.parquet'
    pq.write_table(table, out_file)

  source = ParquetSource(
    filepaths=[str(tmp_path / 'test-*.parquet')], pseudo_shuffle=True, sample_size=20, seed=42
  )
  source.setup()
  manifest = source.load_to_parquet(str(tmp_path), task_id=None)
  items = retrieve_parquet_rows(tmp_path, manifest)
  assert len(items) == 20


def test_validation() -> None:
  with pytest.raises(ValidationError, match='filepaths must be non-empty'):
    ParquetSource(filepaths=[])

  with pytest.raises(ValidationError, match='sample_size must be greater than 0'):
    ParquetSource(filepaths=['gs://lilac/test.parquet'], sample_size=0)


def test_parquet_with_map_of_string_to_array_of_ints(tmp_path: pathlib.Path) -> None:
  create_project_and_set_env(str(tmp_path))
  arrow_schema = pa.schema(
    {
      'name': pa.map_(pa.string(), pa.list_(pa.int32())),
    }
  )
  table = pa.Table.from_pylist(
    [
      {'name': {'key1': [1, 2, 3]}},
      {'name': {'key2': [2, 3]}},
      {'name': {'key1': [3, 4], 'key2': [5, 6]}},
    ],
    schema=arrow_schema,
  )
  out_file = os.path.join(tmp_path, 'test.parquet')
  pq.write_table(table, out_file)
  config = DatasetConfig(
    namespace='local', name='parquet_with_map', source=ParquetSource(filepaths=[out_file])
  )
  dataset = create_dataset(config)

  expected_map_dtype = MapType(key_type=STRING, value_field=Field(repeated_field=field('int32')))
  assert dataset.manifest().data_schema == schema(
    {
      'name': field(
        dtype=expected_map_dtype,
        # All the unique keys of the map got extracted as sub-fields in the Lilac schema.
        fields={'key1': ['int32'], 'key2': ['int32']},
      )
    }
  )
  df = dataset.to_pandas()
  expected_df = pd.DataFrame(
    [
      {'name': {'key': ['key1'], 'value': [[1, 2, 3]]}},
      {'name': {'key': ['key2'], 'value': [[2, 3]]}},
      {'name': {'key': ['key1', 'key2'], 'value': [[3, 4], [5, 6]]}},
    ]
  )
  pd.testing.assert_frame_equal(df, expected_df)



"""Tests for dataset.compute_signal() when signals are chained."""

import re
from typing import ClassVar, Iterable, Iterator, List, Optional, cast

import numpy as np
import pytest
from pytest_mock import MockerFixture
from typing_extensions import override

from ..schema import (
  EMBEDDING_KEY,
  Field,
  Item,
  RichData,
  SignalInputType,
  SpanVector,
  field,
  lilac_embedding,
  schema,
  span,
)
from ..signal import (
  TextEmbeddingSignal,
  TextSignal,
  VectorSignal,
  clear_signal_registry,
  register_signal,
)
from ..source import clear_source_registry, register_source
from .dataset import DatasetManifest
from .dataset_test_utils import (
  TEST_DATASET_NAME,
  TEST_NAMESPACE,
  TestDataMaker,
  TestSource,
  enriched_item,
)

SIMPLE_ITEMS: list[Item] = [
  {'str': 'a', 'int': 1, 'bool': False, 'float': 3.0},
  {'str': 'b', 'int': 2, 'bool': True, 'float': 2.0},
  {'str': 'b', 'int': 2, 'bool': True, 'float': 1.0},
]

EMBEDDINGS: list[tuple[str, list[float]]] = [
  ('hello.', [1.0, 0.0, 0.0]),
  ('hello2.', [1.0, 1.0, 0.0]),
  ('hello world.', [1.0, 1.0, 1.0]),
  ('hello world2.', [2.0, 1.0, 1.0]),
]

STR_EMBEDDINGS: dict[str, list[float]] = {text: embedding for text, embedding in EMBEDDINGS}


class TestSplitter(TextSignal):
  """Split documents into sentence by splitting on period."""

  name: ClassVar[str] = 'test_splitter'

  @override
  def fields(self) -> Field:
    return field(fields=['string_span'])

  @override
  def compute(self, data: Iterable[RichData]) -> Iterator[Item]:
    for text in data:
      if not isinstance(text, str):
        raise ValueError(f'Expected text to be a string, got {type(text)} instead.')
      sentences = [f'{sentence.strip()}.' for sentence in text.split('.') if sentence]
      yield [
        span(text.index(sentence), text.index(sentence) + len(sentence)) for sentence in sentences
      ]


class TestEmbedding(TextEmbeddingSignal):
  """A test embed function."""

  name: ClassVar[str] = 'test_embedding'

  @override
  def compute(self, data: Iterable[RichData]) -> Iterator[Item]:
    """Call the embedding function."""
    for example in data:
      yield [lilac_embedding(0, len(example), np.array(STR_EMBEDDINGS[cast(str, example)]))]


class TestEmbeddingSumSignal(VectorSignal):
  """Sums the embeddings to return a single floating point value."""

  name: ClassVar[str] = 'test_embedding_sum'
  input_type: ClassVar[SignalInputType] = SignalInputType.TEXT

  @override
  def fields(self) -> Field:
    return field('float32')

  @override
  def vector_compute(
    self, all_vector_spans: Iterable[list[SpanVector]]
  ) -> Iterator[Optional[Item]]:
    # The signal just sums the values of the embedding.
    for vector_spans in all_vector_spans:
      yield float(vector_spans[0]['vector'].sum())


@pytest.fixture(scope='module', autouse=True)
def setup_teardown() -> Iterable[None]:
  # Setup.
  register_source(TestSource)
  register_signal(TestSplitter)
  register_signal(TestEmbedding)
  register_signal(TestEmbeddingSumSignal)
  register_signal(NamedEntity)
  # Unit test runs.
  yield
  # Teardown.
  clear_source_registry()
  clear_signal_registry()


def test_manual_embedding_signal(make_test_data: TestDataMaker, mocker: MockerFixture) -> None:
  dataset = make_test_data([{'text': 'hello.'}, {'text': 'hello2.'}])

  embed_mock = mocker.spy(TestEmbedding, 'compute')
  dataset.compute_embedding('test_embedding', 'text')
  embedding_sum_signal = TestEmbeddingSumSignal(embedding='test_embedding')
  dataset.compute_signal(embedding_sum_signal, 'text')

  # Make sure the embedding signal is not called twice.
  assert embed_mock.call_count == 1

  assert dataset.manifest() == DatasetManifest(
    namespace=TEST_NAMESPACE,
    dataset_name=TEST_DATASET_NAME,
    data_schema=schema(
      {
        'text': field(
          'string',
          fields={
            'test_embedding_sum(embedding=test_embedding)': field(
              'float32', signal=embedding_sum_signal.model_dump(exclude_none=True)
            ),
            'test_embedding': field(
              signal=TestEmbedding().model_dump(exclude_none=True),
              fields=[field('string_span', fields={EMBEDDING_KEY: 'embedding'})],
            ),
          },
        )
      }
    ),
    num_items=2,
    source=TestSource(),
  )

  result = dataset.select_rows(combine_columns=True)
  expected_result = [
    {'text': enriched_item('hello.', {'test_embedding_sum(embedding=test_embedding)': 1.0})},
    {'text': enriched_item('hello2.', {'test_embedding_sum(embedding=test_embedding)': 2.0})},
  ]
  assert list(result) == expected_result


def test_missing_embedding_signal(make_test_data: TestDataMaker, mocker: MockerFixture) -> None:
  dataset = make_test_data([{'text': 'hello.'}, {'text': 'hello2.'}])

  # The embedding is missing for 'text'.
  embedding_sum_signal = TestEmbeddingSumSignal(embedding=TestEmbedding.name)
  with pytest.raises(ValueError, match="No embedding found for path \\('text',\\)"):
    dataset.compute_signal(embedding_sum_signal, 'text')


ENTITY_REGEX = r'[A-Za-z]+@[A-Za-z]+'


class NamedEntity(TextSignal):
  """Find special entities."""

  name: ClassVar[str] = 'entity'

  @override
  def fields(self) -> Field:
    return field(fields=['string_span'])

  @override
  def compute(self, data: Iterable[RichData]) -> Iterator[Optional[List[Item]]]:
    for text in data:
      if not isinstance(text, str):
        yield None
        continue
      yield [span(m.start(0), m.end(0)) for m in re.finditer(ENTITY_REGEX, text)]



"""Tests for the dataset_format module."""


from .data.dataset_test_utils import TestDataMaker


def test_infer_no_format(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(
    [
      {'extra': 2, 'question': 'hello?'},
      {'extra': 1, 'question': 'anybody?'},
    ]
  )

  assert dataset.manifest().dataset_format is None



"""Defines the concept and the concept models."""
import dataclasses
from enum import Enum
from typing import Callable, Optional, Union

import numpy as np
from joblib import Parallel, delayed
from pydantic import BaseModel, field_validator
from sklearn.base import clone
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import precision_recall_curve, roc_auc_score
from sklearn.model_selection import KFold

from ..embeddings.embedding import get_embed_fn
from ..signal import TextEmbeddingSignal, get_signal_cls
from ..utils import DebugTimer

LOCAL_CONCEPT_NAMESPACE = 'local'

# The maximum number of cross-validation models to train.
MAX_NUM_CROSS_VAL_MODELS = 15
# The β weight to use for the F-beta score: https://scikit-learn.org/stable/modules/generated/sklearn.metrics.fbeta_score.html
# β = 0.5 means we value precision 2x as much as recall.
# β = 2 means we value recall 2x as much as precision.
F_BETA_WEIGHT = 0.5


class ConceptType(str, Enum):
  """Enum holding the concept type."""

  TEXT = 'text'
  IMAGE = 'image'

  def __repr__(self) -> str:
    return self.value


class ExampleOrigin(BaseModel):
  """The origin of an example."""

  # The namespace that holds the dataset.
  dataset_namespace: str

  # The name of the dataset.
  dataset_name: str

  # The id of row in the dataset that the example was added from.
  dataset_row_id: str


DraftId = str
DRAFT_MAIN = 'main'


class ExampleIn(BaseModel):
  """An example in a concept without the id (used for adding new examples)."""

  label: bool
  text: Optional[str] = None
  img: Optional[bytes] = None
  origin: Optional[ExampleOrigin] = None
  # The name of the draft to put the example in. If None, puts it in the main draft.
  draft: Optional[DraftId] = DRAFT_MAIN

  @field_validator('text')
  @classmethod
  def parse_text(cls, text: str) -> str:
    """Fixes surrogate errors in text: https://github.com/ijl/orjson/blob/master/README.md#str ."""
    return text.encode('utf-8', 'replace').decode('utf-8')


class Example(ExampleIn):
  """A single example in a concept used for training a concept model."""

  id: str


class ConceptMetadata(BaseModel):
  """Metadata associated with a concept."""

  # True if the concept is publicly visible.
  is_public: bool = False
  tags: list[str] = []
  description: Optional[str] = None


class Concept(BaseModel):
  """A concept is a collection of examples."""

  # The namespace of the concept.
  namespace: str
  # The name of the concept.
  concept_name: str
  # The type of the data format that this concept represents.
  type: ConceptType
  data: dict[str, Example]
  version: int = 0

  metadata: Optional[ConceptMetadata] = None

  def drafts(self) -> list[DraftId]:
    """Gets all the drafts for the concept."""
    drafts: set[DraftId] = set([DRAFT_MAIN])  # Always return the main draft.
    for example in self.data.values():
      if example.draft:
        drafts.add(example.draft)
    return list(sorted(drafts))


class OverallScore(str, Enum):
  """Enum holding the overall score."""

  NOT_GOOD = 'not_good'
  OK = 'ok'
  GOOD = 'good'
  VERY_GOOD = 'very_good'
  GREAT = 'great'


def _get_overall_score(f1_score: float) -> OverallScore:
  if f1_score < 0.5:
    return OverallScore.NOT_GOOD
  if f1_score < 0.8:
    return OverallScore.OK
  if f1_score < 0.9:
    return OverallScore.GOOD
  if f1_score < 0.95:
    return OverallScore.VERY_GOOD
  return OverallScore.GREAT


class ConceptMetrics(BaseModel):
  """Metrics for a concept."""

  # The average F1 score for the concept computed using cross validation.
  f1: float
  precision: float
  recall: float
  roc_auc: float
  overall: OverallScore


@dataclasses.dataclass
class LogisticEmbeddingModel:
  """A model that uses logistic regression with embeddings."""

  _metrics: Optional[ConceptMetrics] = None
  _threshold: float = 0.5

  def __post_init__(self) -> None:
    # See `notebooks/Toxicity.ipynb` for an example of training a concept model.
    self._model = LogisticRegression(
      class_weight='balanced', C=30, tol=1e-5, warm_start=True, max_iter=5_000, n_jobs=-1
    )

  def score_embeddings(self, embeddings: np.ndarray) -> np.ndarray:
    """Get the scores for the provided embeddings."""
    y_probs = self._model.predict_proba(embeddings)[:, 1]
    # Map [0, threshold, 1] to [0, 0.5, 1].
    power = np.log(self._threshold) / np.log(0.5)
    return y_probs**power

  def _setup_training(
    self, X_train: np.ndarray, labels: Union[list[bool], np.ndarray]
  ) -> tuple[np.ndarray, np.ndarray]:
    y_train = np.array(labels)
    # Shuffle the data in unison.
    p = np.random.permutation(len(X_train))
    X_train = X_train[p]
    y_train = y_train[p]
    return X_train, y_train

  def fit(self, embeddings: np.ndarray, labels: list[bool]) -> None:
    """Fit the model to the provided embeddings and labels."""
    label_set = set(labels)
    if len(label_set) < 2:
      dim = embeddings.shape[1]
      random_vector = np.random.randn(dim).astype(np.float32)
      random_vector /= np.linalg.norm(random_vector)
      embeddings = np.vstack([embeddings, random_vector])
      labels.append(False if True in label_set else True)

    if len(labels) != len(embeddings):
      raise ValueError(
        f'Length of embeddings ({len(embeddings)}) must match length of labels ({len(labels)})'
      )
    X_train, y_train = self._setup_training(embeddings, labels)
    self._model.fit(X_train, y_train)
    self._metrics, self._threshold = self._compute_metrics(embeddings, labels)

  def _compute_metrics(
    self, embeddings: np.ndarray, labels: list[bool]
  ) -> tuple[Optional[ConceptMetrics], float]:
    """Return the concept metrics."""
    labels_np = np.array(labels)
    n_splits = min(len(labels_np), MAX_NUM_CROSS_VAL_MODELS)
    fold = KFold(n_splits, shuffle=True, random_state=42)

    def _fit_and_score(
      model: LogisticRegression,
      X_train: np.ndarray,
      y_train: np.ndarray,
      X_test: np.ndarray,
      y_test: np.ndarray,
    ) -> tuple[np.ndarray, np.ndarray]:
      if len(set(y_train)) < 2:
        return np.array([]), np.array([])
      model.fit(X_train, y_train)
      y_pred = model.predict_proba(X_test)[:, 1]
      return y_test, y_pred

    # Compute the metrics for each validation fold in parallel.
    jobs: list[Callable] = []
    for train_index, test_index in fold.split(embeddings):
      X_train, y_train = embeddings[train_index], labels_np[train_index]
      X_train, y_train = self._setup_training(X_train, y_train)
      X_test, y_test = embeddings[test_index], labels_np[test_index]
      model = clone(self._model)
      jobs.append(delayed(_fit_and_score)(model, X_train, y_train, X_test, y_test))
    results = Parallel(n_jobs=-1)(jobs)

    y_test = np.concatenate([y_test for y_test, _ in results], axis=0)
    y_pred = np.concatenate([y_pred for _, y_pred in results], axis=0)
    if len(set(y_test)) < 2:
      return None, 0.5
    roc_auc_val = roc_auc_score(y_test, y_pred)
    precision, recall, thresholds = precision_recall_curve(y_test, y_pred)
    numerator = (1 + F_BETA_WEIGHT**2) * precision * recall
    denom = (F_BETA_WEIGHT**2 * precision) + recall
    f1_scores = np.divide(numerator, denom, out=np.zeros_like(denom), where=(denom != 0))
    max_f1: float = np.max(f1_scores)
    max_f1_index = np.argmax(f1_scores)
    max_f1_thresh: float = thresholds[max_f1_index]
    max_f1_prec: float = precision[max_f1_index]
    max_f1_recall: float = recall[max_f1_index]
    metrics = ConceptMetrics(
      f1=max_f1,
      precision=max_f1_prec,
      recall=max_f1_recall,
      roc_auc=float(roc_auc_val),
      overall=_get_overall_score(max_f1),
    )
    return metrics, max_f1_thresh


def draft_examples(concept: Concept, draft: DraftId) -> dict[str, Example]:
  """Get the examples in the provided draft by overriding the main draft."""
  draft_examples: dict[str, dict[str, Example]] = {}
  for id, example in concept.data.items():
    draft_examples.setdefault(example.draft or DRAFT_MAIN, {})[example.id] = example

  if draft == DRAFT_MAIN:
    return draft_examples.get(DRAFT_MAIN, {})

  if draft not in draft_examples:
    raise ValueError(
      f'Draft {draft} not found in concept. Found drafts: {list(draft_examples.keys())}'
    )

  # Map the text of the draft to its id so we can dedupe with main.
  draft_text_ids = {example.text: id for id, example in draft_examples[draft].items()}

  # Write each of examples from main to the draft examples only if the text does not appear in the
  # draft.
  for id, example in draft_examples[DRAFT_MAIN].items():
    if example.text not in draft_text_ids:
      draft_examples[draft][id] = example

  return draft_examples[draft]


@dataclasses.dataclass
class ConceptModel:
  """A concept model. Stores all concept model drafts and manages syncing."""

  # The concept that this model is for.
  namespace: str
  concept_name: str

  # The name of the embedding for this model.
  embedding_name: str
  version: int = 0

  batch_size = 4096

  # The following fields are excluded from JSON serialization, but still pickle-able.
  # Maps a concept id to the embeddings.
  _embeddings: dict[str, np.ndarray] = dataclasses.field(default_factory=dict)
  _logistic_models: dict[DraftId, LogisticEmbeddingModel] = dataclasses.field(default_factory=dict)

  def get_metrics(self) -> Optional[ConceptMetrics]:
    """Return the metrics for this model."""
    return self._get_logistic_model(DRAFT_MAIN)._metrics

  def score_embeddings(self, draft: DraftId, embeddings: np.ndarray) -> np.ndarray:
    """Get the scores for the provided embeddings."""
    return self._get_logistic_model(draft).score_embeddings(embeddings)

  def coef(self, draft: DraftId = DRAFT_MAIN) -> np.ndarray:
    """Get the coefficients of the underlying ML model."""
    return self._get_logistic_model(draft)._model.coef_.reshape(-1)

  def _get_logistic_model(self, draft: DraftId = DRAFT_MAIN) -> LogisticEmbeddingModel:
    """Get the logistic model for the provided draft."""
    if draft not in self._logistic_models:
      self._logistic_models[draft] = LogisticEmbeddingModel()
    return self._logistic_models[draft]

  def sync(self, concept: Concept) -> bool:
    """Update the model with the latest labeled concept data."""
    if concept.version == self.version:
      # The model is up to date.
      return False

    concept_path = f'{self.namespace}/{self.concept_name}/' f'{self.embedding_name}'
    with DebugTimer(f'Computing embeddings for "{concept_path}"'):
      self._compute_embeddings(concept)

    # Fit each of the drafts, sort by draft name for deterministic behavior.
    for draft in concept.drafts():
      examples = draft_examples(concept, draft)
      embeddings = np.array([self._embeddings[id] for id in examples.keys()])
      labels = [example.label for example in examples.values()]
      model = self._get_logistic_model(draft)
      with DebugTimer(f'Fitting model for "{concept_path}"'):
        model.fit(embeddings, labels)

    # Synchronize the model version with the concept version.
    self.version = concept.version

    return True

  def _compute_embeddings(self, concept: Concept) -> None:
    signal_cls = get_signal_cls(self.embedding_name)
    if not signal_cls:
      raise ValueError(f'Embedding signal "{self.embedding_name}" not found in the registry.')
    embedding_signal = signal_cls()
    if not isinstance(embedding_signal, TextEmbeddingSignal):
      raise ValueError(
        f'Only text embedding signals are currently supported for concepts. '
        f'"{self.embedding_name}" is a {type(embedding_signal)}.'
      )

    embed_fn = get_embed_fn(self.embedding_name, split=False)
    concept_embeddings: dict[str, np.ndarray] = {}

    examples = concept.data.items()
    if not examples:
      raise ValueError(f'Cannot sync concept "{concept.concept_name}". It has no examples.')

    # Compute the embeddings for the examples with cache miss.
    texts_of_missing_embeddings: dict[str, str] = {}
    for id, example in examples:
      if id in self._embeddings:
        # Cache hit.
        concept_embeddings[id] = self._embeddings[id]
      else:
        # Cache miss.
        # TODO(smilkov): Support images.
        texts_of_missing_embeddings[id] = example.text or ''

    missing_ids = texts_of_missing_embeddings.keys()
    missing_embeddings = embed_fn(list(texts_of_missing_embeddings.values()))

    for id, (embedding, *_) in zip(missing_ids, missing_embeddings):
      concept_embeddings[id] = embedding['vector'] / np.linalg.norm(embedding['vector'])
    self._embeddings = concept_embeddings



"""Loads files from GitHub using the LlamaIndex GithubRepositoryReader."""
import os
from typing import ClassVar, Iterable, Optional

from pydantic import Field, field_serializer
from typing_extensions import override

from ..schema import Item
from ..source import Source, SourceSchema
from .llama_index_docs_source import LlamaIndexDocsSource

# We currently don't support images or videos, so we filter them out to reduce the load time.
IGNORE_MEDIA_EXTENSIONS = [
  '.png',
  '.jpg',
  '.jpeg',
  '.gif',
  '.mp4',
  '.mov',
  '.avi',
  '.PNG',
  '.JPG',
  '.JPEG',
  '.GIF',
  '.MP4',
  '.MOV',
  '.AVI',
]


class GithubSource(Source):
  """GitHub source code loader

  Loads source code from GitHub repositories using the LlamaIndex GithubRepositoryReader.

  Each file becomes a separate row.

  The following extensions are automatically ignored as Lilac does not yet support media:

  .png, .jpg, .jpeg, .gif, .mp4, .mov, .avi
  """  # noqa: D415, D400

  name: ClassVar[str] = 'github'

  repo: str = Field(description='The GitHub repository to load from. Format: <owner>/<repo>.')
  branch: Optional[str] = Field(
    default='main', description='The branch to load from. Defaults to the main branch.'
  )
  ignore_directories: Optional[list[str]] = Field(
    default=None,
    description='A list of directories to ignore. Can only be used if filter_directories '
    'is not specified.',
  )
  ignore_file_extensions: Optional[list[str]] = Field(
    default=None,
    description='A list of file extensions to ignore. Can only be used if filter_file_extensions '
    'is not specified.',
  )

  github_token: Optional[str] = Field(
    default=None,
    description='The GitHub token to use for authentication. If not specified, '
    'uses the `GITHUB_TOKEN` environment variable.',
  )

  @field_serializer('github_token')
  def scrub_github_token(self, github_token: str) -> str:
    """Scrubs the github token so it isn't stored on disk."""
    del github_token
    return ''

  _llama_index_docs_source: LlamaIndexDocsSource

  @override
  def setup(self) -> None:
    try:
      from llama_index import download_loader

    except ImportError:
      raise ImportError(
        'Could not import dependencies for the "github" source. '
        'Please install with pip install lilac[github]'
      )

    try:
      from llama_hub.github_repo import GithubClient, GithubRepositoryReader

    except ImportError:
      raise ImportError(
        'Could not import dependencies for the "github" source. '
        'Please install with pip install lilac[github]'
      )

    download_loader('GithubRepositoryReader')

    github_token = os.getenv('GITHUB_TOKEN', self.github_token)

    if not github_token:
      raise ValueError(
        'Environment variable `GITHUB_TOKEN` is not set and the github_token arg is not set.'
      )
    github_client = GithubClient(github_token)

    owner, repo = self.repo.split('/')

    loader = GithubRepositoryReader(
      github_client=github_client,
      owner=owner,
      repo=repo,
      filter_directories=(self.ignore_directories, GithubRepositoryReader.FilterType.EXCLUDE)
      if self.ignore_directories
      else None,
      filter_file_extensions=(
        (self.ignore_file_extensions or []) + IGNORE_MEDIA_EXTENSIONS,
        GithubRepositoryReader.FilterType.EXCLUDE,
      ),
      verbose=True,
      concurrent_requests=10,
    )
    docs = loader.load_data(branch=self.branch)

    self._llama_index_docs_source = LlamaIndexDocsSource(docs)
    self._llama_index_docs_source.setup()

  @override
  def source_schema(self) -> SourceSchema:
    """Return the source schema."""
    return self._llama_index_docs_source.source_schema()

  @override
  def yield_items(self) -> Iterable[Item]:
    """Read from GitHub."""
    return self._llama_index_docs_source.yield_items()



"""Cohere embeddings."""
from typing import TYPE_CHECKING, ClassVar, Optional

import numpy as np
from typing_extensions import override

from ..env import env
from ..schema import Item
from ..signal import TextEmbeddingSignal
from ..splitters.spacy_splitter import clustering_spacy_chunker
from ..tasks import TaskExecutionType
from .embedding import chunked_compute_embedding

if TYPE_CHECKING:
  from cohere import Client

COHERE_EMBED_MODEL = 'embed-english-light-v3.0'


class Cohere(TextEmbeddingSignal):
  """Computes embeddings using Cohere's embedding API.

  <br>**Important**: This will send data to an external server!

  <br>To use this signal, you must get a Cohere API key from
  [cohere.com/embed](https://cohere.com/embed) and add it to your .env.local.

  <br>For details on pricing, see: https://cohere.com/pricing.
  """

  name: ClassVar[str] = 'cohere'
  display_name: ClassVar[str] = 'Cohere Embeddings'
  local_batch_size: ClassVar[int] = 96
  local_parallelism: ClassVar[int] = 10
  local_strategy: ClassVar[TaskExecutionType] = 'threads'

  _model: 'Client'

  @override
  def setup(self) -> None:
    """Validate that the api key and python package exists in environment."""
    api_key = env('COHERE_API_KEY')
    if not api_key:
      raise ValueError('`COHERE_API_KEY` environment variable not set.')
    try:
      import cohere

      self._model = cohere.Client(api_key, max_retries=10)
    except ImportError:
      raise ImportError(
        'Could not import the "cohere" python package. '
        'Please install it with `pip install cohere`.'
      )

  @override
  def compute(self, docs: list[str]) -> list[Optional[Item]]:
    """Compute embeddings for the given documents."""
    cohere_input_type = 'search_document' if self.embed_input_type == 'document' else 'search_query'

    def _embed_fn(docs: list[str]) -> list[np.ndarray]:
      return self._model.embed(
        docs, truncate='END', model=COHERE_EMBED_MODEL, input_type=cohere_input_type
      ).embeddings

    return chunked_compute_embedding(
      _embed_fn, docs, self.local_batch_size, chunker=clustering_spacy_chunker
    )
